from http import HTTPStatus
from unittest import mock

from django_otp.plugins.otp_static.models import StaticDevice
from django_otp.plugins.otp_totp.models import TOTPDevice

from django.core import mail
from django.test import TestCase
from django.urls import reverse

from hidp.otp.devices import reset_static_tokens
from hidp.otp.forms import VerifyTOTPForm
from hidp.test.factories import otp_factories, user_factories


class TestOTPOverview(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp_management:manage"))
        self.assertRedirects(
            response, f"/login/?next={reverse('hidp_otp_management:manage')}"
        )

    def test_get_otp_overview_without_devices(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:manage"))
        self.assertContains(response, "Authenticator app: not configured")
        self.assertContains(response, "Recovery codes: not configured")

    def test_get_otp_overview_with_devices(self):
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:manage"))
        self.assertContains(response, "Authenticator app: configured")
        self.assertContains(response, "Recovery codes: configured")

    def test_get_otp_overview_with_unconfirmed_devices(self):
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=False)
        otp_factories.StaticDeviceFactory(user=self.user, confirmed=False)
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:manage"))
        self.assertContains(response, "Authenticator app: not configured")
        self.assertContains(response, "Recovery codes: not configured")


class TestOTPDisable(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp_management:disable"))
        self.assertRedirects(
            response, f"/login/?next={reverse('hidp_otp_management:disable')}"
        )

    def test_get_otp_disable(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:disable"))
        self.assertTemplateUsed(response, "hidp/otp/disable.html")

    @mock.patch.object(VerifyTOTPForm, "clean_otp", return_value=None, autospec=True)
    def test_post_otp_disable(self, mock_clean_otp):
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        static_device = otp_factories.StaticDeviceFactory(
            user=self.user, confirmed=True
        )
        otp_factories.StaticTokenFactory.create_batch(10, device=static_device)
        self.client.force_login(self.user)

        form_data = {
            "otp_token": "123456",
        }
        response = self.client.post(reverse("hidp_otp_management:disable"), form_data)
        self.assertRedirects(response, reverse("hidp_otp_management:manage"))
        self.assertFalse(
            self.user.totpdevice_set.exists(),
            msg="Expected the user to have no TOTP devices",
        )
        self.assertFalse(
            self.user.staticdevice_set.exists(),
            msg="Expected the user to have no static devices",
        )

    def test_static_token_not_accepted(self):
        """The *static_token* should not be accepted in the *otp_token* disable form."""
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        static_device = otp_factories.StaticDeviceFactory(
            user=self.user, confirmed=True
        )
        otp_factories.StaticTokenFactory.create_batch(9, device=static_device)
        otp_factories.StaticTokenFactory(token="a1b2c3", device=static_device)
        self.client.force_login(self.user)

        form_data = {
            "otp_token": "a1b2c3",
        }
        response = self.client.post(reverse("hidp_otp_management:disable"), form_data)
        form = response.context["form"]
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid")
        # Check that the error is on the token field
        errors = form.errors.as_data()
        self.assertEqual(errors["__all__"][0].code, "invalid_token")
        self.assertTrue(
            self.user.totpdevice_set.exists(),
            msg="Expected the user to have TOTP devices",
        )
        self.assertTrue(
            self.user.staticdevice_set.exists(),
            msg="Expected the user to have static devices",
        )


class TestOTPDisableWithRecoveryCode(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp_management:disable-recovery-code"))
        self.assertRedirects(
            response,
            f"/login/?next={reverse('hidp_otp_management:disable-recovery-code')}",
        )

    def test_get_otp_disable_with_recovery_code(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:disable-recovery-code"))
        self.assertTemplateUsed(response, "hidp/otp/disable_recovery_code.html")

    def test_post_otp_disable_with_recovery_code(self):
        static_device = otp_factories.StaticDeviceFactory(
            user=self.user, confirmed=True
        )
        otp_factories.StaticTokenFactory.create_batch(9, device=static_device)
        otp_factories.StaticTokenFactory(device=static_device, token="a1b2c3d4")
        self.client.force_login(self.user)

        form_data = {
            "otp_token": "a1b2c3d4",
        }
        with (
            self.assertTemplateUsed("hidp/otp/email/disabled_subject.txt"),
            self.assertTemplateUsed("hidp/otp/email/disabled_body.txt"),
            self.assertTemplateUsed("hidp/otp/email/disabled_body.html")
        ):  # fmt: skip
            response = self.client.post(
                reverse("hidp_otp_management:disable-recovery-code"), form_data
            )
        self.assertRedirects(response, reverse("hidp_otp_management:manage"))
        self.assertFalse(
            self.user.totpdevice_set.exists(),
            msg="Expected the user to have no TOTP devices",
        )
        self.assertFalse(
            self.user.staticdevice_set.exists(),
            msg="Expected the user to have no static devices",
        )

        self.assertEqual(len(mail.outbox), 1, "Expected an email to be sent")
        self.assertEqual(mail.outbox[0].subject, "Two-factor authentication disabled")
        self.assertEqual(mail.outbox[0].to, [self.user.email])


class TestOTPRecoveryCodesView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp_management:recovery-codes"))
        self.assertRedirects(
            response,
            f"/login/?next={reverse('hidp_otp_management:recovery-codes')}",
        )

    def test_no_static_device(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:recovery-codes"))
        self.assertEqual(response.status_code, HTTPStatus.NOT_FOUND)

    def test_get_recovery_codes(self):
        device = otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        reset_static_tokens(device)
        current_tokens = list(device.token_set.values_list("token", flat=True))
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:recovery-codes"))
        self.assertContains(response, "Recovery codes")
        self.assertContains(response, "Generate new codes")
        for token in current_tokens:
            self.assertContains(response, token)

    def test_post_reset_recovery_codes(self):
        device = otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        reset_static_tokens(device)
        current_tokens = set(device.token_set.values_list("token", flat=True))
        self.client.force_login(self.user)
        with (
            self.assertTemplateUsed("hidp/otp/email/recovery_codes_regenerated_subject.txt"),
            self.assertTemplateUsed("hidp/otp/email/recovery_codes_regenerated_body.txt"),
            self.assertTemplateUsed("hidp/otp/email/recovery_codes_regenerated_body.html")
        ):  # fmt: skip
            response = self.client.post(reverse("hidp_otp_management:recovery-codes"))
        self.assertRedirects(response, reverse("hidp_otp_management:recovery-codes"))
        new_tokens = set(device.token_set.values_list("token", flat=True))
        self.assertEqual(new_tokens & current_tokens, set())
        self.assertEqual(len(new_tokens), 10)

        self.assertEqual(len(mail.outbox), 1, "Expected an email to be sent")
        self.assertEqual(mail.outbox[0].subject, "Recovery codes were regenerated")
        self.assertEqual(mail.outbox[0].to, [self.user.email])


class TestOTPSetupView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        """The user must be logged in to set up OTP."""
        response = self.client.get(reverse("hidp_otp_management:setup"))
        self.assertRedirects(
            response, f"/login/?next={reverse('hidp_otp_management:setup')}"
        )

    def test_redirects_to_manage_when_already_setup(self):
        """The user should be redirected to the manage page if OTP is already set up."""
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        self.client.force_login(self.user)
        response = self.client.get(reverse("hidp_otp_management:setup"))
        self.assertRedirects(response, reverse("hidp_otp_management:setup-device-done"))

    def test_redirects_to_next_page_when_already_setup(self):
        """The user should be redirected to the next page after setup."""
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        self.client.force_login(self.user)
        response = self.client.get(
            f"{reverse('hidp_otp_management:setup')}?next=/my-special-page/"
        )
        self.assertRedirects(
            response, "/my-special-page/", fetch_redirect_response=False
        )

    def test_valid_form_confirms_devices(self):
        """A valid form should confirm the TOTP and static devices."""
        self.client.force_login(self.user)
        device = otp_factories.TOTPDeviceFactory(user=self.user, confirmed=False)
        form_data = {
            "otp_token": "123456",
            "confirm_stored_backup_tokens": True,
        }
        with (
            mock.patch.object(device, "verify_token", return_value=True, autospec=True),
            mock.patch("hidp.otp.forms.OTPSetupForm.get_device", return_value=device),
            self.assertTemplateUsed("hidp/otp/email/configured_subject.txt"),
            self.assertTemplateUsed("hidp/otp/email/configured_body.txt"),
            self.assertTemplateUsed("hidp/otp/email/configured_body.html"),
        ):
            response = self.client.post(reverse("hidp_otp_management:setup"), form_data)
        self.assertRedirects(response, reverse("hidp_otp_management:setup-device-done"))

        totp_device = TOTPDevice.objects.get(user=self.user)
        self.assertTrue(totp_device.confirmed, "Expected TOTP device to be confirmed")

        static_device = StaticDevice.objects.get(user=self.user)
        self.assertTrue(
            static_device.confirmed, "Expected static device to be confirmed"
        )

        self.assertEqual(len(mail.outbox), 1, "Expected an email to be sent")
        self.assertEqual(mail.outbox[0].subject, "Two-factor authentication configured")
        self.assertEqual(mail.outbox[0].to, [self.user.email])

    def test_invalid_form_does_not_confirm_devices(self):
        """An invalid form should not confirm the TOTP and static devices."""
        self.client.force_login(self.user)
        form_data = {
            "otp_token": "xxxxxx",  # Invalid token
            "confirm_stored_backup_tokens": True,
        }
        response = self.client.post(reverse("hidp_otp_management:setup"), form_data)
        form = response.context["form"]
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid")
        # Check that the error is on the token field
        errors = form.errors.as_data()
        self.assertEqual(errors["__all__"][0].code, "invalid_token")

        totp_device = TOTPDevice.objects.get(user=self.user)
        self.assertFalse(
            totp_device.confirmed, "Expected TOTP device to be unconfirmed"
        )

        static_device = StaticDevice.objects.get(user=self.user)
        self.assertFalse(
            static_device.confirmed, "Expected static device to be unconfirmed"
        )

    def _setup_up_otp(self, *, next_page=None):
        self.client.force_login(self.user)
        self.assertFalse(
            TOTPDevice.objects.devices_for_user(self.user, confirmed=None).exists()
        )

        setup_url = reverse("hidp_otp_management:setup")
        if next_page:
            setup_url += f"?next={next_page}"

        response = self.client.get(setup_url)
        device = TOTPDevice.objects.devices_for_user(self.user, confirmed=None).get()
        self.assertFalse(
            response.wsgi_request.user.is_verified(), "Expected user to be unverified"
        )
        form_data = {
            "otp_token": "123456",
            "confirm_stored_backup_tokens": True,
        }
        with (
            mock.patch.object(device, "verify_token", return_value=True, autospec=True),
            mock.patch("hidp.otp.forms.OTPSetupForm.get_device", return_value=device),
        ):
            return self.client.post(setup_url, form_data)

    def test_setting_up_otp_verifies_user(self):
        """Setting up OTP successfully should verify the user."""
        response = self._setup_up_otp()
        self.assertRedirects(response, reverse("hidp_otp_management:setup-device-done"))
        self.assertTrue(
            response.wsgi_request.user.is_verified(), "Expected user to be verified"
        )

    def test_setting_up_otp_redirects_to_next_page(self):
        """Setting up OTP successfully should redirect to the next page."""
        response = self._setup_up_otp(next_page="/my-special-page/")
        self.assertRedirects(
            response, "/my-special-page/", fetch_redirect_response=False
        )


class TestOTPVerifyView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp:verify"))
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={reverse('hidp_otp:verify')}",
        )

    def test_valid_form_verifies_user(self):
        device = otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        self.client.force_login(self.user)
        form_data = {"otp_token": "123456"}
        manage_url = reverse("hidp_account_management:manage_account")
        with (
            mock.patch.object(device, "verify_token", return_value=True, autospec=True),
            mock.patch("hidp.otp.forms.VerifyTOTPForm.get_device", return_value=device),
        ):
            response = self.client.post(
                f"{reverse('hidp_otp:verify')}?next={manage_url}", form_data
            )
            self.assertRedirects(response, manage_url)
        self.assertTrue(
            response.wsgi_request.user.is_verified(), "Expected user to be verified"
        )

    def test_invalid_form_does_not_verify_user(self):
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=True)
        self.client.force_login(self.user)
        form_data = {"otp_token": "xxxxxx"}  # Invalid token
        response = self.client.post(reverse("hidp_otp:verify"), form_data)
        form = response.context["form"]
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid")
        # Check that the error is on the token field
        errors = form.errors.as_data()
        self.assertEqual(errors["__all__"][0].code, "invalid_token")
        self.assertFalse(
            response.wsgi_request.user.is_verified(), "Expected user to be unverified"
        )


class TestOTPVerifyWithRecoveryCodeView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()

    def test_requires_login(self):
        response = self.client.get(reverse("hidp_otp:verify-recovery-code"))
        self.assertRedirects(
            response,
            f"{reverse('hidp_accounts:login')}?next={reverse('hidp_otp:verify-recovery-code')}",
        )

    def test_valid_form_verifies_user(self):
        device = otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        otp_factories.StaticTokenFactory(token="123456", device=device)
        self.client.force_login(self.user)
        form_data = {"otp_token": "123456"}
        manage_url = reverse("hidp_account_management:manage_account")
        with (
            self.assertTemplateUsed("hidp/otp/email/recovery_code_used_subject.txt"),
            self.assertTemplateUsed("hidp/otp/email/recovery_code_used_body.txt"),
            self.assertTemplateUsed("hidp/otp/email/recovery_code_used_body.html"),
        ):
            response = self.client.post(
                f"{reverse('hidp_otp:verify-recovery-code')}?next={manage_url}",
                form_data,
            )
        self.assertRedirects(response, manage_url)
        self.assertTrue(
            response.wsgi_request.user.is_verified(), "Expected user to be verified"
        )

        self.assertEqual(len(mail.outbox), 1, "Expected an email to be sent")
        self.assertEqual(mail.outbox[0].subject, "Recovery code used")
        self.assertEqual(mail.outbox[0].to, [self.user.email])

    def test_invalid_form_does_not_verify_user(self):
        device = otp_factories.StaticDeviceFactory(user=self.user, confirmed=True)
        otp_factories.StaticTokenFactory.create_batch(10, device=device)
        self.client.force_login(self.user)
        form_data = {"otp_token": "xxxxxx"}  # Invalid token
        response = self.client.post(reverse("hidp_otp:verify-recovery-code"), form_data)
        form = response.context["form"]
        self.assertFalse(form.is_valid(), msg="Expected form to be invalid")
        # Check that the error is on the token field
        errors = form.errors.as_data()
        self.assertEqual(errors["__all__"][0].code, "invalid_token")
        self.assertFalse(
            response.wsgi_request.user.is_verified(), "Expected user to be unverified"
        )

        self.assertEqual(len(mail.outbox), 0, "Expected no email to be sent")



import logging

from urllib.parse import urlencode

from django.contrib.auth import get_user_model
from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import (
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from django.views import generic

from hidp.utils import is_registration_enabled

from ..accounts import auth as hidp_auth
from ..accounts import email_verification, mailers
from ..config import oidc_clients
from ..csp.decorators import hidp_csp_protection
from ..federated.constants import OIDCError
from ..rate_limit.decorators import rate_limit_strict
from . import forms, tokens
from .models import OpenIdConnection
from .oidc import authorization_code_flow
from .oidc.exceptions import InvalidOIDCStateError, OAuth2Error

logger = logging.getLogger(__name__)

UserModel = get_user_model()


class OIDCMixin:
    callback_pattern = "hidp_oidc_client:callback"

    def dispatch(self, request, *args, **kwargs):
        # Require HTTPS for OIDC requests. This is a security requirement to
        # prevent the authentication request from happening in the clear.
        if not request.is_secure():
            return HttpResponseBadRequest("Insecure request")
        return super().dispatch(request, *args, **kwargs)

    def get_callback_url(self, provider_key):
        return reverse(
            self.callback_pattern,
            kwargs={
                "provider_key": provider_key,
            },
        )


class OIDCContextMixin:
    """Mixin to provide context data for OIDC login providers."""

    oidc_error_messages = {
        OIDCError.ACCOUNT_EXISTS: _(
            "You already have an account with this email address."
            " Please sign in to link your account."
        ),
        OIDCError.REQUEST_EXPIRED: _(
            "The authentication request has expired. Please try again."
        ),
        OIDCError.UNEXPECTED_ERROR: _(
            "An unexpected error occurred during authentication. Please try again."
        ),
        OIDCError.INVALID_TOKEN: _("Expired or invalid token. Please try again."),
        OIDCError.INVALID_CREDENTIALS: _("Login failed. Invalid credentials."),
        OIDCError.REGISTRATION_DISABLED: _("Registration is disabled."),
    }

    @staticmethod
    def _build_provider_url_list(
        providers,
        url_name="hidp_oidc_client:authenticate",
        label=None,
    ):
        return [
            {
                "provider": provider,
                "url": reverse(
                    url_name,
                    kwargs={
                        "provider_key": provider.provider_key,
                    },
                ),
                "label": format_lazy(
                    label or _("Sign in with {provider}"),
                    provider=provider.name,
                ),
            }
            for provider in providers
        ]

    def get_context_data(self, **kwargs):
        oidc_error = self.request.GET.get("oidc_error", None)
        context = {
            "oidc_error": oidc_error,
            "oidc_login_providers": self._build_provider_url_list(
                oidc_clients.get_registered_oidc_clients()
            ),
            "oidc_error_message": self.oidc_error_messages.get(oidc_error, oidc_error),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_strict, name="dispatch")
class OIDCAuthenticationRequestView(
    auth_views.RedirectURLMixin, OIDCMixin, generic.View
):
    """Initiates an OpenID Connect Authorization Code Flow authentication request."""

    # Optionally set extra parameters to include in the authentication request.
    # I.e. to modify the prompt parameter.
    extra_authentication_request_params = None

    http_method_names = [
        "post",
        "options",
    ]

    def post(self, request, *, provider_key):
        """
        Handle the OIDC authentication request.

        Prepare the authentication request parameters, update the session state
        with the required information, and redirect the user to the OpenID Connect
        provider's authorization endpoint.
        """
        return HttpResponseRedirect(
            authorization_code_flow.prepare_authentication_request(
                request,
                client=oidc_clients.get_oidc_client_or_404(provider_key),
                callback_url=self.get_callback_url(provider_key),
                next_url=self.get_redirect_url(),
                **(self.extra_authentication_request_params or {}),
            )
        )


@method_decorator(rate_limit_strict, name="dispatch")
class OIDCAuthenticationCallbackView(OIDCMixin, generic.View):
    """
    Handle the callback from an OIDC authentication request.

    This view is used for both successful and failed authentication attempts.
    """

    http_method_names = [
        "get",
        "options",
    ]

    def get_next_url(  # noqa: PLR6301 (no-self-use)
        self,
        *,
        request,
        provider_key,
        claims,
        user_info,
        redirect_url=None,
    ):
        """Decide which flow the user should be redirected to next."""
        connection = OpenIdConnection.objects.get_by_provider_and_claims(
            provider_key=provider_key,
            issuer_claim=claims["iss"],
            subject_claim=claims["sub"],
        )
        if connection:
            # A connection exists for the given claims. This must be a login attempt.
            token = OIDCLoginView.add_data_to_session(
                request,
                provider_key=provider_key,
                claims=claims,
                user_info=user_info,
            )
            view_name = "hidp_oidc_client:login"
        elif request.user.is_authenticated:
            # `sub` claim does not match an existing user:
            # Display a form allowing the user to link the OIDC account.
            token = OIDCAccountLinkView.add_data_to_session(
                request,
                provider_key=provider_key,
                claims=claims,
                user_info=user_info,
            )
            view_name = "hidp_oidc_management:link_account"
        else:
            # `sub` claim does not match an existing user, and no user is logged in:
            # Check if a user exists for the given email.
            user = UserModel.objects.filter(email__iexact=claims["email"]).first()
            if not user:
                # `sub` and `email` claim do not match an existing user:
                # Redirect the user to the registration page.
                token = OIDCRegistrationView.add_data_to_session(
                    request,
                    provider_key=provider_key,
                    claims=claims,
                    user_info=user_info,
                )
                view_name = "hidp_oidc_client:register"
            else:
                # `sub` claim does not match an existing user, but `email` claim does:
                # Display a message instructing the user to log in to link the accounts.
                return (
                    reverse("hidp_accounts:login")
                    + f"?oidc_error={OIDCError.ACCOUNT_EXISTS}"
                )

        # Prepare the URL parameters for the next view. Drop any None values.
        params = {
            key: value
            for key, value in (
                ("token", token),
                ("next", redirect_url),
            )
            if value is not None
        }
        return reverse(view_name) + f"?{urlencode(params)}"

    def get(self, request, provider_key):
        try:
            _tokens, claims, user_info, next_url = (
                authorization_code_flow.handle_authentication_callback(
                    request,
                    client=oidc_clients.get_oidc_client_or_404(provider_key),
                    callback_url=self.get_callback_url(provider_key),
                )
            )
        except InvalidOIDCStateError:
            # The state parameter in the callback is not present in the session.
            # The user might have tampered with the state parameter, the session
            # might have expired or the authentication request might have expired.
            # Redirect the user to the login page to try again.
            logger.exception("Invalid OIDC state parameter")
            return HttpResponseRedirect(
                reverse("hidp_accounts:login")
                + f"?oidc_error={OIDCError.REQUEST_EXPIRED}"
            )
        except OAuth2Error:
            # One of many things went wrong during the authentication process.
            # Redirect the user to the login page to try again.
            logger.exception("Error during OIDC authentication")
            return HttpResponseRedirect(
                reverse("hidp_accounts:login")
                + f"?oidc_error={OIDCError.UNEXPECTED_ERROR}"
            )

        return HttpResponseRedirect(
            self.get_next_url(
                request=request,
                provider_key=provider_key,
                claims=claims,
                user_info=user_info,
                redirect_url=next_url,
            )
        )


class TokenDataMixin:
    """Mixin to set, retrieve and validate data to/from the session using a token."""

    token_generator = NotImplemented
    invalid_token_redirect_url = reverse_lazy("hidp_accounts:login")

    @classmethod
    def add_data_to_session(cls, request, *, provider_key, claims, user_info):
        token = cls.token_generator.make_token()
        request.session[token] = {
            "provider_key": provider_key,
            "claims": claims,
            "user_info": user_info,
        }
        return token

    def dispatch(self, request, *args, **kwargs):
        self.token = request.GET.get("token")
        valid_token = self.token and self.token_generator.check_token(self.token)
        self.token_data = valid_token and request.session.get(self.token)
        self.provider = (
            oidc_clients.get_oidc_client_or_none(self.token_data["provider_key"])
            if self.token_data
            else None
        )

        if not valid_token or self.provider is None:
            return HttpResponseRedirect(
                self.invalid_token_redirect_url
                + f"?oidc_error={OIDCError.INVALID_TOKEN}"
            )
        return super().dispatch(request, *args, **kwargs)


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_strict, name="dispatch")
class OIDCRegistrationView(
    auth_views.RedirectURLMixin, TokenDataMixin, generic.FormView
):
    """Register a new user using the OIDC provider's claims and user info."""

    token_generator = tokens.OIDCRegistrationTokenGenerator()
    form_class = forms.OIDCRegistrationForm
    template_name = "hidp/federated/registration.html"
    next_page = "/"
    verification_mailer = mailers.EmailVerificationMailer
    invalid_token_redirect_url = reverse_lazy("hidp_accounts:register")

    def dispatch(self, request, *args, **kwargs):
        if not is_registration_enabled():
            # Registration is disabled. Redirect to the login page.
            return HttpResponseRedirect(
                reverse("hidp_accounts:login")
                + f"?oidc_error={OIDCError.REGISTRATION_DISABLED}"
            )
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        return super().get_form_kwargs() | {
            "provider_key": self.token_data["provider_key"],
            "claims": self.token_data["claims"],
            "user_info": self.token_data["user_info"],
        }

    def send_email(self, user):
        """Send the email verification email to the user."""
        self.verification_mailer(
            user,
            base_url=self.request.build_absolute_uri("/"),
            post_verification_redirect=self.get_redirect_url(),
        ).send()

    def form_valid(self, form):
        user = form.save()
        # Remove the token from the session after the form has been saved.
        del self.request.session[self.token]

        # Send the email verification email.
        self.send_email(user)

        # Redirect to the email verification required page.
        return HttpResponseRedirect(
            email_verification.get_email_verification_required_url(
                user, next_url=self.get_redirect_url()
            )
        )


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_strict, name="dispatch")
class OIDCLoginView(auth_views.RedirectURLMixin, TokenDataMixin, generic.FormView):
    """Log in a user using the OIDC provider's claims."""

    token_generator = tokens.OIDCLoginTokenGenerator()
    next_page = "/"
    verification_mailer = mailers.EmailVerificationMailer

    def send_email(self, user):
        """Send the email verification email to the user."""
        self.verification_mailer(
            user,
            base_url=self.request.build_absolute_uri("/"),
            post_verification_redirect=self.get_redirect_url(),
        ).send()

    def get(self, request):
        """
        User has provided valid credentials and is allowed to log in.

        Persist the user and backend in the session and redirect to the
        success URL.

        If the user's email address has not been verified, redirect them
        to the email verification required flow.
        """
        user = hidp_auth.authenticate(
            request,
            provider_key=self.token_data["provider_key"],
            issuer_claim=self.token_data["claims"]["iss"],
            subject_claim=self.token_data["claims"]["sub"],
        )
        if user is None:
            # The user could not be authenticated using the OIDC claims.
            # The account is probably disabled. Just redirect to the login page.
            return HttpResponseRedirect(
                reverse("hidp_accounts:login")
                + f"?oidc_error={OIDCError.INVALID_CREDENTIALS}"
            )

        if user.email_verified:
            # Only log in the user if their email address has been verified.
            hidp_auth.login(self.request, user)
            return HttpResponseRedirect(self.get_success_url())

        # If the user's email address is not yet verified:
        # Send the email verification email.
        self.send_email(user)

        # Then redirect them to the email verification required page.
        return HttpResponseRedirect(
            email_verification.get_email_verification_required_url(
                user, next_url=self.get_redirect_url()
            )
        )


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_strict, name="dispatch")
@method_decorator(login_required, name="dispatch")
class OIDCAccountLinkView(TokenDataMixin, generic.FormView):
    """Link an existing user account to an OIDC account."""

    form_class = forms.OIDCAccountLinkForm
    template_name = "hidp/federated/account_link.html"
    token_generator = tokens.OIDCAccountLinkTokenGenerator()
    invalid_token_redirect_url = reverse_lazy("hidp_oidc_management:linked_services")

    def get_success_url(self):
        return reverse(
            "hidp_oidc_management:link_account_done",
            kwargs={"provider_key": self.provider.provider_key},
        )

    def get_form_kwargs(self):
        return super().get_form_kwargs() | {
            "user": self.request.user,
            "provider_key": self.token_data["provider_key"],
            "claims": self.token_data["claims"],
        }

    def get_context_data(self, **kwargs):
        context = {
            "provider": self.provider,
            "user_email": self.request.user.email,
            "provider_email": self.token_data["claims"]["email"],
            "cancel_url": self.success_url,
        }
        return super().get_context_data() | context | kwargs

    def form_valid(self, form):
        form.save()
        # Remove the token from the session after the form has been saved.
        del self.request.session[self.token]

        return super().form_valid(form)


@method_decorator(hidp_csp_protection, name="dispatch")
class OIDCAccountLinkDoneView(LoginRequiredMixin, generic.TemplateView):
    """Show a success message after linking an OIDC account."""

    template_name = "hidp/federated/account_link_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "provider": oidc_clients.get_oidc_client_or_404(
                self.kwargs["provider_key"]
            ),
            "back_url": reverse("hidp_oidc_management:linked_services"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
class OIDCAccountUnlinkView(LoginRequiredMixin, generic.DeleteView):
    """Unlink an OIDC account from an existing user account."""

    form_class = forms.OIDCAccountUnlinkForm
    template_name = "hidp/federated/account_unlink.html"
    slug_field = "provider_key"
    slug_url_kwarg = "provider_key"

    def dispatch(self, request, *args, provider_key, **kwargs):
        self.provider = oidc_clients.get_oidc_client_or_404(provider_key)
        return super().dispatch(request, *args, **kwargs)

    def get_success_url(self):
        return reverse(
            "hidp_oidc_management:unlink_account_done",
            kwargs={"provider_key": self.provider.provider_key},
        )

    def get_queryset(self):
        return self.request.user.openid_connections

    def get_form_kwargs(self):
        return super().get_form_kwargs() | {
            "user": self.request.user,
            "provider_key": self.provider.provider_key,
        }

    def get_context_data(self, **kwargs):
        context = {
            "provider": self.provider,
            "cancel_url": self.success_url,
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
class OIDCAccountUnlinkDoneView(LoginRequiredMixin, generic.TemplateView):
    """Show a success message after unlinking an OIDC account."""

    template_name = "hidp/federated/account_unlink_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "provider": oidc_clients.get_oidc_client_or_404(
                self.kwargs["provider_key"]
            ),
            "back_url": reverse("hidp_oidc_management:linked_services"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
class OIDCLinkedServicesView(
    LoginRequiredMixin, OIDCContextMixin, generic.TemplateView
):
    """Display the linked services page."""

    template_name = "hidp/federated/linked_services.html"

    def get_context_data(self, **kwargs):
        oidc_linked_provider_keys = self.request.user.openid_connections.values_list(
            "provider_key", flat=True
        )
        # Do not allow the user to unlink the only available login method.
        user = self.request.user
        can_unlink = user.has_usable_password() or user.openid_connections.count() > 1
        linked_provider = oidc_clients.get_oidc_client_or_none(
            self.request.GET.get("success")
        )
        removed_provider = oidc_clients.get_oidc_client_or_none(
            self.request.GET.get("removed")
        )
        context = {
            "successfully_linked_provider": linked_provider,
            "removed_provider": removed_provider,
            "oidc_linked_providers": self._build_provider_url_list(
                (
                    provider
                    for provider in oidc_clients.get_registered_oidc_clients()
                    if provider.provider_key in oidc_linked_provider_keys
                ),
                url_name="hidp_oidc_management:unlink_account",
                label=_("Unlink from {provider}"),
            ),
            "oidc_available_providers": self._build_provider_url_list(
                (
                    provider
                    for provider in oidc_clients.get_registered_oidc_clients()
                    if provider.provider_key not in oidc_linked_provider_keys
                ),
                label=_("Link with {provider}"),
            ),
            "can_unlink": can_unlink,
            "set_password_url": reverse("hidp_account_management:set_password"),
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs



from django_otp.forms import (
    OTPAuthenticationFormMixin as DjangoOTPAuthenticationFormMixin,
)
from django_otp.plugins.otp_static.models import StaticDevice
from django_otp.plugins.otp_totp.models import TOTPDevice

from django import forms
from django.utils.translation import gettext_lazy as _

from hidp.otp.devices import get_device_for_user


class OTPAuthenticationFormMixin(DjangoOTPAuthenticationFormMixin):
    # Override/copy the error messages to be able to translate them in HIdP
    otp_error_messages = DjangoOTPAuthenticationFormMixin.otp_error_messages | {
        "invalid_token": _(
            "Invalid token. Please make sure you have entered it correctly."
        ),
    }


class OTPVerifyFormBase(OTPAuthenticationFormMixin, forms.Form):
    # The device class to use for verification, e.g. TOTPDevice or StaticDevice.
    device_class = None

    @staticmethod
    def create_otp_token_field(label):
        """
        Create a form field for the OTP token.

        Args:
            label: Field label

        Returns:
            Configured CharField
        """
        attrs = {
            "autocomplete": "one-time-code",
            "inputmode": "numeric",
            "pattern": "[0-9]*",
        }

        return forms.CharField(
            widget=forms.TextInput(attrs=attrs),
            label=label,
            max_length=6,
        )

    @staticmethod
    def create_recovery_code_field(label):
        """
        Create a form field for the recovery code.

        Args:
            label: Field label

        Returns:
            Configured CharField
        """
        attrs = {
            "autocomplete": "off",
        }

        return forms.CharField(
            widget=forms.TextInput(attrs=attrs),
            label=label,
            max_length=8,
        )

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.user = user

    def _chosen_device(self, user):
        return self.get_device(user)

    def get_device(self, user):
        return get_device_for_user(user, self.device_class)

    def clean(self):
        super().clean()

        self.clean_otp(self.user)

        return self.cleaned_data


class VerifyTOTPForm(OTPVerifyFormBase):
    """
    A form used to verify a TOTP token from an Authenticator App.

    This form is used to verify a TOTP token entered by the user. It will verify the
    token against the user's confirmed TOTP device.
    """

    device_class = TOTPDevice
    otp_token = OTPVerifyFormBase.create_otp_token_field(
        label=_("Enter the code from the app"),
    )


class VerifyStaticTokenForm(OTPVerifyFormBase):
    """
    A form used to verify a static token from a list of recovery codes.

    This form is used to verify a static token entered by the user. It will verify
    the token against the user's confirmed Static device.
    """

    device_class = StaticDevice
    otp_token = OTPVerifyFormBase.create_recovery_code_field(
        label=_("Enter a recovery code"),
    )


class OTPSetupForm(OTPVerifyFormBase):
    otp_token = OTPVerifyFormBase.create_otp_token_field(
        label=_("Enter the code from the app"),
    )
    confirm_stored_backup_tokens = forms.BooleanField(
        required=True,
        label=_("I have stored my backup codes in a safe place"),
        help_text=_(
            "You can use these codes to log in if you lose access to your device"
        ),
    )

    def get_device(self, user):
        """Hard-wire the unconfirmed device to verify against."""
        return self.device

    def __init__(self, *args, user, device, backup_device, **kwargs):
        super().__init__(user, *args, **kwargs)
        self.device = device
        self.backup_device = backup_device

    def save(self):
        # Mark the devices as confirmed
        self.device.confirmed = True
        self.device.save(update_fields=["confirmed"])
        self.backup_device.confirmed = True
        self.backup_device.save(update_fields=["confirmed"])



from django.contrib import auth as django_auth
from django.contrib.auth.base_user import AbstractBaseUser


def authenticate(request, **credentials):
    """
    Attempts to authenticate a user using all configured authentication backends.

    Note that this does **not** log the user in. It only verifies the credentials.

    Iterates over each backend in `settings.AUTHENTICATION_BACKENDS` until it finds
    one that accepts the credentials, leading to three possible outcomes:

    1. `None` is returned:
       The backend could not verify the credentials. The credentials may be
       invalid, or the backend is unable to produce a user for some other reason.

    2. A `User` object is returned:
       The credentials are valid and the user is allowed to log in.

    3. `PermissionDenied` is raised:
       The backend was able to verify the credentials, but the user is not
       allowed to log in.

    In the first scenario, the next backend is tried. In the other two
    scenarios, the iteration is stopped.

    Returns the authenticated user, annotated with the path of the backend that
    authenticated the user as `user.backend`, if successful.

    Returns `None` if the credentials are invalid, or access is denied,
    and sends the `django.contrib.auth.user_login_failed` signal.
    """
    # Wrap Django's authenticate, without altering its behavior, to add
    # a detailed docstring and provide a consistent interface for the
    # `hidp.accounts.auth` module.
    return django_auth.authenticate(request=request, **credentials)


def login(request, user, backend=None):
    """
    Logs in the given user, persisting the user and backend in the request's session.

    Assumes that sessions are enabled, e.g. by the
    `django.contrib.sessions.middleware.SessionMiddleware` middleware.

    It is recommended to only call this function with the user returned by
    a successful call to `authenticate`.

    Two scenarios are possible:

    1. The request session does not have a user:
       The existing anonymous session data is retained. The session key is
       cycled to avoid session fixation attacks.

    2. The request session user is different from given user:
       The session is flushed to avoid reusing another user's session.
       This logs out the existing user and rotates the session key.

    Does **not** verify that the user is allowed to log in. Might result
    in inconsistent behaviour when user argument is not the return value of
    `authenticate`.

    The `backend` argument is optional. If not provided, the user's `backend`
    attribute is used. The user may not have a `backend` attribute if it was
    not retrieved using `authenticate`.

    If the user does not have a `backend` attribute, and there is only one
    authentication backend configured, that backend is used. Otherwise,
    a `ValueError` is raised.

    If no exception is raised, the CSRF token is rotated and the
    `django.contrib.auth.user_logged_in` signal is sent.

    Django listens to this signal to update the `last_login` field of the
    user object (if the field exists) with the current date and time.
    """
    # Be explicit about the expected type of the user argument. Do not handle
    # None values, unlike Django, to avoid unexpected behavior.
    # See also:
    # https://code.djangoproject.com/ticket/35530#comment:1
    if not isinstance(user, AbstractBaseUser):
        raise TypeError(f"{type(user).__name__!r} does not extend AbstractBaseUser")
    django_auth.login(request, user, backend=backend)


def logout(request):
    """
    Logs out the user, regardless of whether a user is logged in.

    Removes all session data, including the user and the CSRF token,
    and cycles the session key.

    Sends the `django.contrib.auth.user_logged_out` signal **before**
    the session is flushed. This allows listeners to access the user
    and the request before the session is reset.
    """
    # Wrap Django's logout, without altering its behavior, to add
    # a detailed docstring and provide a consistent interface for the
    # `hidp.accounts.auth` module.
    django_auth.logout(request)



"""
OAuth2 Provider URLs.

Provides the URL patterns for OAuth2 and OpenID Connect (OIDC) endpoints.

Include this module in the root URL configuration:

    from hidp.accounts import oidc_provider_urls

    urlpatterns = [
        path("o/", include(oidc_provider_urls)),
    ]

This module uses the `oauth2_provider` namespace for these URLs.

Include this namespace when reversing URLs, for example:

    reverse("oauth2_provider:authorize")

Does **not** include application management views. Applications can be managed
using the Django Admin interface.
"""

from oauth2_provider import views as oauth2_views

from django.urls import path

from ..rate_limit.decorators import rate_limit_default, rate_limit_strict
from . import views

app_name = "oauth2_provider"

base_urlpatterns = [
    path(
        "authorize/",
        rate_limit_strict(views.AuthorizationView.as_view()),
        name="authorize",
    ),
    path(
        "token/",
        rate_limit_strict(oauth2_views.TokenView.as_view()),
        name="token",
    ),
    path(
        "revoke_token/",
        rate_limit_default(oauth2_views.RevokeTokenView.as_view()),
        name="revoke-token",
    ),
    path(
        "introspect/",
        rate_limit_default(oauth2_views.IntrospectTokenView.as_view()),
        name="introspect",
    ),
]

oidc_urlpatterns = [
    path(
        ".well-known/openid-configuration",
        rate_limit_default(oauth2_views.ConnectDiscoveryInfoView.as_view()),
        name="oidc-connect-discovery-info",
    ),
    path(
        ".well-known/jwks.json",
        rate_limit_default(oauth2_views.JwksInfoView.as_view()),
        name="jwks-info",
    ),
    path(
        "userinfo/",
        rate_limit_default(oauth2_views.UserInfoView.as_view()),
        name="user-info",
    ),
    path(
        "logout/",
        rate_limit_default(views.RPInitiatedLogoutView.as_view()),
        name="rp-initiated-logout",
    ),
]

urlpatterns = base_urlpatterns + oidc_urlpatterns



"""
Generate time-sortable UUIDs (version 7) (RFC 9562).

Provides a `uuid7` function that's either directly imported from Python's `uuid` module
(if it's available) or an implementation based on a pull request to add it to CPython.
"""

import os
import uuid

if hasattr(uuid, "uuid7"):
    # Future Python versions will (hopefully) have this function built-in
    # Issue: https://github.com/python/cpython/issues/89083
    uuid7 = uuid.uuid7
else:
    # Taken from the CPython pull request:
    # * https://github.com/python/cpython/pull/121119
    # * Commit (2024-06-28T09:40:44Z)
    # * https://github.com/python/cpython/blob/ef85b200602ab00c23ef158813fa57076f561cfd/Lib/uuid.py#L752-L815
    # Modifications:
    # * Added noqa comments to supress ruff warnings
    # * Minor formatting changes
    # * Use int constructor argument instead of private _from_int method,
    #   the latter does not (yet) exist

    _RFC_4122_VERSION_7_FLAGS = (7 << 76) | (0x8000 << 48)
    _last_timestamp_v7 = None
    _last_counter_v7 = 0  # 42-bit counter

    def uuid7():
        """
        Generate a UUID from a Unix timestamp in milliseconds and random bits.

        UUIDv7 objects feature monotonicity within a millisecond.
        """
        # --- 48 ---   -- 4 --   --- 12 ---   -- 2 --   --- 30 ---   - 32 -
        # unix_ts_ms | version | counter_hi | variant | counter_lo | random
        #
        # 'counter = counter_hi | counter_lo' is a 42-bit counter constructed
        # with Method 1 of RFC 9562, §6.2, and its MSB is set to 0.
        #
        # 'random' is a 32-bit random value regenerated for every new UUID.
        #
        # If multiple UUIDs are generated within the same millisecond, the LSB
        # of 'counter' is incremented by 1. When overflowing, the timestamp is
        # advanced and the counter is reset to a random 42-bit integer with MSB
        # set to 0.

        def get_counter_and_tail():
            rand = int.from_bytes(os.urandom(10))
            # 42-bit counter with MSB set to 0
            counter = (rand >> 32) & 0x1FF_FFFF_FFFF
            # 32-bit random data
            tail = rand & 0xFFFF_FFFF
            return counter, tail

        global _last_timestamp_v7, _last_counter_v7  # noqa: PLW0603

        import time  # noqa: PLC0415

        nanoseconds = time.time_ns()
        timestamp_ms = nanoseconds // 1_000_000

        if _last_timestamp_v7 is None or timestamp_ms > _last_timestamp_v7:
            counter, tail = get_counter_and_tail()
        else:
            if timestamp_ms < _last_timestamp_v7:
                timestamp_ms = _last_timestamp_v7 + 1
            # advance the 42-bit counter
            counter = _last_counter_v7 + 1
            if counter > 0x3FF_FFFF_FFFF:  # noqa: PLR2004
                timestamp_ms += 1  # advance the 48-bit timestamp
                counter, tail = get_counter_and_tail()
            else:
                tail = int.from_bytes(os.urandom(4))

        _last_timestamp_v7 = timestamp_ms
        _last_counter_v7 = counter

        unix_ts_ms = timestamp_ms & 0xFFFF_FFFF_FFFF
        counter_msbs = counter >> 30
        counter_hi = counter_msbs & 0x0FFF  # keep 12 bits and clear variant bits
        counter_lo = counter & 0x3FFF_FFFF  # keep 30 bits and clear version bits

        int_uuid_7 = unix_ts_ms << 80
        int_uuid_7 |= counter_hi << 64
        int_uuid_7 |= counter_lo << 32
        int_uuid_7 |= tail & 0xFFFF_FFFF
        # by construction, the variant and version bits are already cleared
        int_uuid_7 |= _RFC_4122_VERSION_7_FLAGS

        return uuid.UUID(int=int_uuid_7)



import django_otp
import segno

from django_otp.plugins.otp_static.models import StaticDevice
from django_otp.plugins.otp_totp.models import TOTPDevice

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import RedirectURLMixin
from django.db import transaction
from django.forms import Form
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse, reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, FormView, TemplateView

from hidp.csp.decorators import hidp_csp_protection
from hidp.otp.devices import (
    STATIC_DEVICE_NAME,
    TOTP_DEVICE_NAME,
    get_or_create_devices,
    reset_static_tokens,
)
from hidp.otp.forms import OTPSetupForm, VerifyStaticTokenForm, VerifyTOTPForm
from hidp.rate_limit.decorators import rate_limit_default

from .decorators import otp_exempt
from .mailers import (
    OTPConfiguredMailer,
    OTPDisabledMailer,
    RecoveryCodesRegeneratedMailer,
    RecoveryCodeUsedMailer,
)


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(login_required, name="dispatch")
class OTPOverviewView(TemplateView):
    template_name = "hidp/otp/overview.html"

    def get_context_data(self, **kwargs):
        context = {
            "totp_devices": TOTPDevice.objects.devices_for_user(
                self.request.user, confirmed=True
            ),
            "static_devices": StaticDevice.objects.devices_for_user(
                self.request.user, confirmed=True
            ),
            "TOTP_DEVICE_NAME": TOTP_DEVICE_NAME,
            "STATIC_DEVICE_NAME": STATIC_DEVICE_NAME,
            "back_url": reverse("hidp_account_management:manage_account"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(login_required, name="dispatch")
class OTPDisableView(FormView):
    """
    View to disable OTP for a user.

    This view will delete all OTP devices for a user, effectively disabling OTP for
    that user. Disabling requires the user to be logged in and to provide a valid OTP
    token.
    """

    template_name = "hidp/otp/disable.html"
    form_class = VerifyTOTPForm
    success_url = reverse_lazy("hidp_otp_management:manage")

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_otp_management:manage"),
        }
        return super().get_context_data() | context | kwargs

    def get_form_kwargs(self):
        context = {
            "user": self.request.user,
        }
        return super().get_form_kwargs() | context

    @transaction.atomic
    def form_valid(self, form):
        for device in django_otp.devices_for_user(self.request.user):
            device.delete()

        self.send_mail()

        return super().form_valid(form)

    def send_mail(self):
        base_url = self.request.build_absolute_uri("/")

        OTPDisabledMailer(self.request.user, base_url=base_url).send()


class OTPDisableRecoveryCodesView(OTPDisableView):
    """
    View to disable OTP for a user using a recovery code.

    This view will delete all OTP devices for a user, effectively disabling OTP for
    that user. Disabling requires the user to be logged in and to provide a valid
    recovery code.
    """

    template_name = "hidp/otp/disable_recovery_code.html"
    form_class = VerifyStaticTokenForm


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(login_required, name="dispatch")
class OTPRecoveryCodesView(DetailView, FormView):
    """
    View for managing recovery codes.

    This view allows the user to reset their recovery codes.
    """

    template_name = "hidp/otp/recovery_codes.html"
    context_object_name = "device"
    form_class = Form
    success_url = reverse_lazy("hidp_otp_management:recovery-codes")

    def get_object(self, queryset=None):
        return get_object_or_404(
            StaticDevice.objects.devices_for_user(self.request.user)
        )

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_otp_management:manage"),
            "recovery_codes": "\n".join(
                self.object.token_set.values_list("token", flat=True)
            ),
        }
        return super().get_context_data() | context | kwargs

    def form_valid(self, form):
        reset_static_tokens(self.get_object())

        self.send_mail()

        return super().form_valid(form)

    def send_mail(self):
        base_url = self.request.build_absolute_uri("/")

        RecoveryCodesRegeneratedMailer(self.request.user, base_url=base_url).send()


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(login_required, name="dispatch")
@method_decorator(otp_exempt, name="dispatch")
class OTPSetupDeviceView(RedirectURLMixin, FormView):
    """
    View for setting up a new OTP device.

    This view will create a new TOTP and Static device in unconfirmed state for the user
    if they don't already have them. The user must verify the TOTP device by entering a
    valid token and declare that they have saved the recovery codes before the devices
    are confirmed.
    """

    form_class = OTPSetupForm
    next_page = reverse_lazy("hidp_otp_management:setup-device-done")
    template_name = "hidp/otp/setup_device.html"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.device = None
        self.user = None

    def dispatch(self, request, *args, **kwargs):
        self.user = request.user

        # If the user already has a confirmed TOTP device, redirect to the manage page
        if TOTPDevice.objects.devices_for_user(self.user, confirmed=True).exists():
            return HttpResponseRedirect(self.get_success_url())

        self.device, self.backup_device = get_or_create_devices(self.user)

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(
            {
                "user": self.user,
                "device": self.device,
                "backup_device": self.backup_device,
            }
        )
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        return context | {
            "title": _("Set up two-factor authentication"),
            "device": self.device,
            "backup_device": self.backup_device,
            "qrcode": segno.make(self.device.config_url).svg_data_uri(border=0),
            "config_url": self.device.config_url,
            "recovery_codes": "\n".join(
                self.backup_device.token_set.values_list("token", flat=True)
            ),
            "back_url": reverse("hidp_otp_management:manage"),
        }

    def form_valid(self, form):
        form.save()
        django_otp.login(self.request, self.device)

        self.send_mail()

        return super().form_valid(form)

    def send_mail(self):
        base_url = self.request.build_absolute_uri("/")

        OTPConfiguredMailer(self.user, base_url=base_url).send()


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(login_required, name="dispatch")
class OTPSetupDeviceDoneView(TemplateView):
    template_name = "hidp/otp/setup_device_done.html"

    def get_context_data(self, **kwargs):
        context = {
            "back_url": reverse("hidp_otp_management:manage"),
        }
        return super().get_context_data() | context | kwargs


@method_decorator(hidp_csp_protection, name="dispatch")
@method_decorator(rate_limit_default, name="dispatch")
@method_decorator(login_required, name="dispatch")
@method_decorator(otp_exempt, name="dispatch")
class VerifyOTPBase(RedirectURLMixin, FormView):
    next_page = settings.LOGIN_REDIRECT_URL

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        # Persist the OTP device in the session
        django_otp.login(self.request, self.request.user.otp_device)
        return super().form_valid(form)


class VerifyTOTPView(VerifyOTPBase):
    template_name = "hidp/otp/verify.html"
    form_class = VerifyTOTPForm


class VerifyRecoveryCodeView(VerifyOTPBase):
    template_name = "hidp/otp/verify_recovery_code.html"
    form_class = VerifyStaticTokenForm

    def form_valid(self, form):
        result = super().form_valid(form)

        self.send_mail()

        return result

    def send_mail(self):
        """Notify the user that a recovery code was used."""
        base_url = self.request.build_absolute_uri("/")
        RecoveryCodeUsedMailer(self.request.user, base_url=base_url).send()



from django.http import HttpResponse


def rate_limited(request, exception):
    return HttpResponse(
        content=(
            "Sorry, you have made too many requests to the server."
            " Please try again later."
        ),
        status=429,
    )



from django.conf import settings
from django.db import models
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from ..compat.uuid7 import uuid7
from ..config import oidc_clients


class OpenIdConnectionQuerySet(models.QuerySet):
    def get_by_provider_and_claims(self, provider_key, *, issuer_claim, subject_claim):
        """
        Get an OpenID connection by provider key and claims.

        Prefetches the associated user along with the query.

        Args:
            provider_key (str): The provider key.
            issuer_claim (str): The issuer claim.
            subject_claim (str): The subject claim.

        Returns:
            OpenIdConnection: The OpenID connection if found, otherwise None.
        """
        return (
            self.select_related("user")
            .filter(
                provider_key=provider_key,
                issuer_claim=issuer_claim,
                subject_claim=subject_claim,
            )
            .first()
        )


class OpenIdConnection(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid7, editable=False)

    # The associated user in HIdP
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="openid_connections",
    )

    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)

    # Short identifier for the identity provider (e.g. google, microsoft, etc.).
    provider_key = models.CharField(max_length=100)
    # The issuer of the identity token (Microsoft has tenant specific issuers).
    issuer_claim = models.CharField(max_length=255)
    # Unique identifier for the user at the identity provider (i.e. the "sub" claim)
    # Guaranteed to be unique together with the issuer_claim.
    subject_claim = models.CharField(max_length=255)

    # Manager
    objects = OpenIdConnectionQuerySet.as_manager()

    class Meta:
        # The sub (subject) and iss (issuer) Claims, used together,
        # are the only Claims that an RP can rely upon as a stable identifier
        # for the End-User, since the sub Claim MUST be locally unique and
        # never reassigned within the Issuer for a particular End-User [...].
        # Therefore, the only guaranteed unique identifier for a given End-User
        # is the combination of the iss Claim and the sub Claim.
        # https://openid.net/specs/openid-connect-basic-1_0.html#ClaimStability
        unique_together = (
            "provider_key",
            "issuer_claim",
            "subject_claim",
        )
        verbose_name = _("OpenID connection")
        verbose_name_plural = _("OpenID connections")

    def __str__(self):
        provider = oidc_clients.get_oidc_client_or_none(self.provider_key)
        provider_name = (
            provider.name
            if provider
            else format_lazy(
                _("Unknown provider: {provider_key}"),
                provider_key=self.provider_key,
            )
        )
        return f"{provider_name} ({self.subject_claim})"



import factory

from django_otp.plugins.otp_static.models import StaticToken
from factory.django import DjangoModelFactory


class TOTPDeviceFactory(DjangoModelFactory):
    class Meta:
        model = "otp_totp.TOTPDevice"

    name = "Authenticator app"
    confirmed = False


class StaticDeviceFactory(DjangoModelFactory):
    class Meta:
        model = "otp_static.StaticDevice"

    name = "Recovery codes"
    confirmed = False


class StaticTokenFactory(DjangoModelFactory):
    class Meta:
        model = StaticToken

    device = factory.SubFactory(StaticDeviceFactory)
    token = factory.LazyFunction(StaticToken.random_token)



import logging
import os
import warnings

from pathlib import Path

from hidp import config as hidp_config

# Enable all warnings
warnings.resetwarnings()
# Warn only once per module
warnings.simplefilter("module")
# Redirect warnings output to the logging system
logging.captureWarnings(capture=True)

# Disable all log output, except warnings
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {"class": "logging.StreamHandler"},
        "null": {"class": "logging.NullHandler"},
    },
    "loggers": {
        "": {
            "handlers": ["null"],
        },
        "py.warnings": {
            "handlers": ["console"],
            "level": "WARNING",
        },
    },
}

# Repository root directory
BASE_DIR = Path(__file__).resolve().parent.parent.parent.parent

# Shared var directory (for logs, cache, etc.)
VAR_DIR = BASE_DIR / "var"

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "rest_framework",
    "oauth2_provider",
    "hidp",
    "hidp.accounts",
    "hidp.api",
    "hidp.csp",
    "hidp.federated",
    "hidp.oidc_provider",
    "hidp.otp",
    "django_otp",
    "django_otp.plugins.otp_static",
    "django_otp.plugins.otp_totp",
    # Custom user model
    "tests.custom_user",
    # Custom makemessages command
    "tests.translations",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "hidp.rate_limit.middleware.RateLimitMiddleware",
    "hidp.oidc_provider.middleware.UiLocalesMiddleware",
    "django_otp.middleware.OTPMiddleware",
]

USE_TZ = True

AUTH_USER_MODEL = "custom_user.CustomUser"

# Custom authentication backend
AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "hidp.federated.auth.backends.OIDCModelBackend",
]

# Login and logout settings
LOGIN_URL = "hidp_accounts:login"
LOGIN_REDIRECT_URL = "/"
LOGOUT_REDIRECT_URL = "/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
            ],
        },
    }
]

OAUTH2_PROVIDER = hidp_config.get_oauth2_provider_settings(
    OIDC_RSA_PRIVATE_KEY=(VAR_DIR / "oidc.key").read_text(),
)

# Test key
SECRET_KEY = "secret-key-only-for-testing"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres" if "CI" in os.environ else "test_hidp",
        "USER": "postgres",
        "PASSWORD": "postgres",
        "HOST": "localhost" if "CI" in os.environ else "postgres",
    }
}

ALLOWED_HOSTS = ["*"]

# Disable caches
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    }
}

# Enable unsafe but fast hashing, we're just testing anyway
PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.MD5PasswordHasher",
]

ROOT_URLCONF = "hidp.config.urls"

# Help Django find and update HIdP's message catalogs
LOCALE_PATHS = [
    BASE_DIR / "packages" / "hidp" / "hidp" / "locale",
]



from django.test import TestCase

from hidp.test.factories import user_factories
from tests.custom_user.models import CustomUser


class TestUserManager(TestCase):
    def test_create_user_requires_email(self):
        with self.assertRaises(ValueError) as cm:
            CustomUser.objects.create_user(email=None)
        self.assertEqual("User must have an email address", str(cm.exception))

    def test_creates_normal_users(self):
        """
        Create a normal user with the given email address.

        The email address should be normalized, the password should be hashed,
        and the user should not be a staff member or a superuser.
        """
        user = CustomUser.objects.create_user(
            email="info@EXAMPLE.COM", password="P@ssw0rd!"
        )
        self.assertEqual(user.email, "info@example.com")
        self.assertTrue(
            user.has_usable_password(), msg="Expected user to have a usable password"
        )
        self.assertTrue(
            user.check_password("P@ssw0rd!"), msg="Expected password to match"
        )
        self.assertFalse(user.is_staff, msg="Expected is_staff to be False")
        self.assertFalse(user.is_superuser, msg="Expected is_superuser to be False")

    def test_creates_superusers(self):
        superuser = CustomUser.objects.create_superuser(
            email="info@EXAMPLE.COM", password="P@ssw0rd!"
        )
        self.assertEqual(superuser.email, "info@example.com")
        self.assertTrue(
            superuser.has_usable_password(),
            msg="Expected user to have a usable password",
        )
        self.assertTrue(
            superuser.check_password("P@ssw0rd!"), msg="Expected password to match"
        )
        self.assertTrue(superuser.is_staff, msg="Expected is_staff to be True")
        self.assertTrue(superuser.is_superuser, msg="Expected is_superuser to be True")

    def test_email_unverified(self):
        """Return a queryset of users that have not verified their email address."""
        unverified_user = user_factories.UserFactory()
        user_factories.VerifiedUserFactory()
        user_factories.SuperUserFactory()
        self.assertQuerySetEqual(
            CustomUser.objects.email_unverified(),
            [repr(unverified_user)],
            transform=repr,
            ordered=False,
        )

    def test_email_verified(self):
        """Return a queryset of users that have verified their email address."""
        user_factories.UserFactory()
        verified_user = user_factories.VerifiedUserFactory()
        superuser = user_factories.SuperUserFactory()
        self.assertQuerySetEqual(
            CustomUser.objects.email_verified(),
            [repr(verified_user), repr(superuser)],
            transform=repr,
            ordered=False,
        )



import base64
import hashlib

from hidp.federated.providers.base import OIDCClient


class ExampleOIDCClient(OIDCClient):
    # A perfectly valid OIDC client, with all the required attributes
    # and a valid provider key. It just doesn't work because it's an example.
    provider_key = "example"
    issuer = "https://example.com"
    authorization_endpoint = "https://example.com/auth"
    token_endpoint = "https://example.com/token"
    userinfo_endpoint = "https://example.com/userinfo"
    jwks_uri = "https://example.com/jwks"


def code_challenge_from_code_verifier(cove_verifier):
    # Re-implements code_challenge generation for testing purposes.
    return (
        base64.urlsafe_b64encode(hashlib.sha256(cove_verifier.encode("ascii")).digest())
        .decode("ascii")
        .rstrip("=")
    )



"""Provide the basic interface, and some common functionality, for all providers."""

import urllib.parse


def _valid_provider_key(provider_key):
    """The provider key is valid when URL-encoding it does not change the value."""
    return urllib.parse.quote(provider_key) == provider_key


def _valid_callback_base_url(callback_base_url):
    """
    Determine the validity of a callback base url.

    The callback base url is valid when the scheme is available and set to https,
    a netloc is present, the path is empty or points to the root, and all
    other parts are empty.
    """
    scheme, netloc, path, query, fragment = urllib.parse.urlsplit(callback_base_url)
    path = path.rstrip("/")  # Remove trailing slash so "/" becomes ""
    return scheme == "https" and netloc and not (path or query or fragment)


def _valid_endpoint(endpoint):
    """Communication with endpoints MUST utilize TLS."""
    # In order to prevent man-in-the-middle attacks, the authorization
    # server MUST require the use of TLS with [...] for any request sent
    # to the authorization and token endpoints.
    # https://datatracker.ietf.org/doc/html/rfc6749#section-10.9
    return endpoint.startswith("https://")


class OIDCClient:
    """
    Base class for OpenID Connect clients.

    Needs to be overridden and configured, according to the provider's specifications.

    Attributes:
        provider_key (``str``):
            Provider key, used to identify the provider in the application.

            This should be unique, descriptive, url-safe and, preferably, lowercase.

        name (``str``, `optional`):
            Provider name, used for display purposes. Will default to the capitalized
            provider key if not set.

        issuer (``str``):
            OpenID Connect issuer URL.

        authorization_endpoint (``str``):
            OpenID Connect authorization endpoint URL.

        token_endpoint (``str``):
            OpenID Connect token endpoint URL.

        userinfo_endpoint (``str``):
            OpenID Connect userinfo endpoint URL.

        jwks_uri (``str``):
            OpenID Connect JWKS URI.

        has_pkce_support (``boolean``):
            Whether the provider supports PKCE (Proof Key for Code Exchange).

            Note: Only set to True if the provider supports S256 as the code
            challenge method.

        client_id (``str``):
            Provider assigned client ID.

        client_secret (``str``, `optional`):
            Provider assigned client secret, if required for token exchange.

        callback_base_url (``str``, `optional`):
            Alternative base URL to use instead of the one of the request when
            constructing the callback URL.

    """

    provider_key = NotImplemented
    name = None

    # OpenID Connect configuration, can usually be extracted from the
    # provider's discovery document, commonly found at:
    # https://<provider>/.well-known/openid-configuration
    issuer = NotImplemented
    authorization_endpoint = NotImplemented
    token_endpoint = NotImplemented
    userinfo_endpoint = NotImplemented
    jwks_uri = NotImplemented

    # Note: Only set to True if the provider supports S256 as the code challenge method.
    has_pkce_support = True
    # Note: If set to False the client secret will be sent in the request body.
    #       This is not spec compliant but required for some providers (e.g. LinkedIn).
    has_basic_auth_support = True

    client_id = None
    client_secret = None
    callback_base_url = None
    # Note: Most providers should support these scopes. If not, (e.g. Facebook) override
    #       the attribute in the subclass.
    scope = "openid email profile"

    def __init__(self, *, client_id, client_secret=None, callback_base_url=None):
        """
        Initialize the OpenID Connect client.

        Arguments:
            client_id (``str``):
                The client ID provided by the OpenID Connect provider.

            client_secret (``str``, `optional`):
                The client secret provided by the OpenID Connect provider.
                Leave as None if the provider does not require a client secret.

            callback_base_url (``str``, `optional`):
                Alternative base URL to use instead of the one of the request
                when constructing the callback URL.

                Some providers require the callback URL to be on a public domain,
                which may not be the case during development.
                Leave as None to use the request domain.

        """
        # Only allow instantiation if all required attributes are set.
        if any(
            value is NotImplemented
            for value in (
                self.provider_key,
                self.issuer,
                self.authorization_endpoint,
                self.token_endpoint,
                self.userinfo_endpoint,
                self.jwks_uri,
            )
        ):
            raise NotImplementedError(
                f"{self.__class__.__name__!r} misses (some of) the required attributes."
            )

        if not _valid_provider_key(self.provider_key):
            raise ValueError(
                f"'{self.__class__.__name__}.provider_key' is not URL-safe:"
                f" {self.provider_key!r}"
            )

        for endpoint in (
            self.authorization_endpoint,
            self.token_endpoint,
            self.userinfo_endpoint,
            self.jwks_uri,
        ):
            if not _valid_endpoint(endpoint):
                raise ValueError(
                    f"All endpoints must use TLS (https): {endpoint!r} does not."
                )

        if not self.has_pkce_support and client_secret is None:
            raise ValueError(
                f"Please provide a client secret."
                f" {self.__class__.__name__!r} declares it does not support PKCE,"
                " which means a client secret must be required for token exchange."
            )

        if callback_base_url is not None and not _valid_callback_base_url(
            callback_base_url
        ):
            raise ValueError(
                f"Invalid callback base url: {callback_base_url!r}."
                f" Should be in the form of 'https://<netloc>'"
                " (path, querystring and/or fragment are not allowed)."
            )

        # Validation passed, initialize the client.
        if self.name is None:
            self.name = self.provider_key.capitalize()

        self.client_id = client_id
        self.client_secret = client_secret
        self.callback_base_url = callback_base_url

    def get_issuer(self, *, claims):
        """
        Return the expected value of the 'iss' claim in the ID token.

        Only override this method if absolutely necessary.

        Arguments:
            claims (``dict``):
                The claims from the ID token.

        Returns:
            ``str``:
                The expected issuer URL based on the claims.
        """
        # Some providers (like Microsoft) have a different issuer
        # for each tenant. This method allows to return the expected
        # issuer based on the claims.
        return self.issuer

    @staticmethod
    def normalize_userinfo(*, userinfo):
        """
        Normalize the data from the userinfo endpoint to the standard format.

        This method should only be overridden if absolutely necessary.

        Arguments:
            userinfo (``dict``):
                The raw data from the userinfo endpoint.

        Returns:
            ``dict``:
                The normalized userinfo data.
        """
        # Some providers return user information in a different format.
        #
        # For example, Facebook uses 'id' instead of 'sub' and 'first_name' &
        # 'last_name' instead of 'given_name' & 'family_name'.
        #
        # Remap the keys to the standard claims if necessary.
        return userinfo



import warnings

from django.conf import settings
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from .config import oidc_clients


def is_registration_enabled():
    if hasattr(settings, "REGISTRATION_ENABLED"):
        # Return the set value
        return settings.REGISTRATION_ENABLED
    else:
        # Preserve the default behavior, as it was before this release by returning
        # True if the setting is not set. This prevents breaking changes for
        # existing installations where the setting was not set and the default
        # behavior was to allow registration.
        warnings.warn(
            "The default value of the REGISTRATION_ENABLED setting will change "
            "from True to False in a future version of HIdP. Set REGISTRATION_ENABLED "
            "to True to maintain the current situation or to False to silence this "
            "warning.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return True


def get_account_management_links(user):
    """
    Get a list of account management links for the given user.

    This function returns a list of dictionaries representing navigation
    links for common account-related actions. These include editing
    account details, changing or setting a password, managing linked
    OpenID Connect (OIDC) services, and configuring two-factor authentication (2FA),
    depending on the user's capabilities and the enabled features in the application.

    Args:
        user: The currently authenticated user.

    Returns:
        list[dict]: A list of link dictionaries with 'url' and 'text' keys.
    """
    if not user.is_authenticated:
        return []

    links = [
        {
            "url": reverse("hidp_account_management:edit_account"),
            "text": _("Edit account"),
        },
        {
            "url": reverse("hidp_account_management:email_change_request"),
            "text": _("Change email address"),
        },
    ]

    if user.has_usable_password():
        links.append(
            {
                "url": reverse("hidp_account_management:change_password"),
                "text": _("Change password"),
            },
        )
    else:
        links.append(
            {
                "url": reverse("hidp_account_management:set_password"),
                "text": _("Set a password"),
            },
        )

    if oidc_clients.get_registered_oidc_clients():
        links.append(
            {
                "url": reverse("hidp_oidc_management:linked_services"),
                "text": _("Linked services"),
            },
        )

    if "hidp.otp" in settings.INSTALLED_APPS:
        links.append(
            {
                "url": reverse("hidp_otp_management:manage"),
                "text": _("Two-factor authentication"),
            },
        )

    return links



from unittest import mock

from django.http import HttpResponse
from django.test import RequestFactory, TestCase, override_settings
from django.utils import translation

from hidp.oidc_provider.middleware import UiLocalesMiddleware


@override_settings(LANGUAGES=[("en", "English"), ("fr", "French")])
class TestUiLocalesMiddleware(TestCase):
    client_class = RequestFactory

    def setUp(self):
        self.get_response = mock.Mock(return_value=HttpResponse())
        self.middleware = UiLocalesMiddleware(self.get_response)

    def test_no_preference(self):
        """Do nothing if there is no language preference."""
        active_language = translation.get_language()
        response = self.middleware(self.client.get("/"))
        self.assertEqual(
            response,
            self.get_response.return_value,
        )
        self.assertEqual(active_language, translation.get_language())
        self.assertNotIn(
            "django_language",
            response.cookies,
        )

    def test_sets_cookie_for_supported_langauge(self):
        """Sets the language cookie if the param contains a supported language."""
        request = self.client.get("/?ui_locales=fr")
        response = self.middleware(request)
        # Sets the language cookie.
        self.assertEqual(
            response.cookies["django_language"].value,
            "fr",
        )

    def test_picks_first_supported_language(self):
        """Picks the first supported language if multiple are provided."""
        request = self.client.get("/?ui_locales=de fr-be en")
        response = self.middleware(request)
        # Sets the language cookie.
        self.assertEqual(
            response.cookies["django_language"].value,
            # 'de' is not supported, 'fr-be' is a variant of 'fr', so 'fr' is picked.
            "fr",
        )

    def test_ignores_unsupported_language(self):
        """Does not set the language cookie if the language is not supported."""
        request = self.client.get("/?ui_locales=de")
        response = self.middleware(request)
        # Does not set the language cookie.
        self.assertNotIn(
            "django_language",
            response.cookies,
        )



"""Functions to handle the OIDC Authorization Code Flow, with optional PKCE support."""

# 2.1.  Code Flow
#
# The Code Flow consists of the following steps:
#
# 1. Client prepares an Authentication Request containing the desired
#    request parameters.
# 2. Client sends the request to the Authorization Server.
# 3. Authorization Server authenticates the End-User.
# 4. Authorization Server obtains End-User Consent/Authorization.
# 5. Authorization Server sends the End-User back to the Client with code.
# 6. Client sends the code to the Token Endpoint to receive an Access Token
#    and ID Token in the response.
# 7. Client validates the tokens and retrieves the End-User's Subject Identifier.
#
# https://openid.net/specs/openid-connect-basic-1_0.html#CodeFlow

import base64
import collections
import hashlib
import json
import secrets
import time

from datetime import timedelta
from urllib.parse import urlencode, urljoin, urlsplit

import requests

from jwcrypto import jwt

from ..constants import OIDC_STATES_SESSION_KEY
from . import jwks
from .exceptions import InvalidOIDCStateError, OAuth2Error, OIDCError

# Maximum number of concurrent state entries (authentication requests)
# that can be stored in the session. It should be enough to handle
# concurrent requests, but within reason.
_OIDC_MAX_STATE_ENTRIES = 25

# How long the state should be stored in the session.
#
# This is the maximum amount of time between the initial authentication
# request and the callback from the OpenID Connect provider.
#
# This should be a middle ground between security and usability.
#
# State should be kept for long enough to allow for the following scenarios:
# - Network latency (the user might be on a slow connection).
# - The user might be going through a password reset flow at the provider.
#
# However, state should not be kept for too long, as:
# - The user might simply get distracted.
# - The user might close the browser and come back later.
# - The user might just abandon the authentication process.
#
# So, at some point, the state should be considered stale and removed.
_OIDC_STATE_TTL = timedelta(minutes=15).total_seconds()


def _build_absolute_uri(request, client, redirect_uri):
    """
    Builds an absolute URI for the redirect URI.

    Uses the current request base URL, unless the client defines a
    callback base URL.
    """
    return urljoin(
        client.callback_base_url or request.build_absolute_uri("/"),
        redirect_uri,
    )


def _clamp_state_entries(states):
    """
    Clamps the number of state entries to the maximum allowed.

    If the number of state entries exceeds the maximum, the oldest
    entries are removed.
    """
    return dict(
        sorted(
            states.items(), key=lambda item: item[1].get("created_at", 0), reverse=True
        )[:_OIDC_MAX_STATE_ENTRIES]
    )


def _add_state_to_session(request, state_key, *, next_url=None):
    """Adds a state to the session, to be used in the authentication response."""
    # Multiple concurrent authentication requests might be happening at the
    # same time. A dictionary is used to store the state for each request.
    states = request.session.get(OIDC_STATES_SESSION_KEY, {})
    states[state_key] = {
        # Allow the state to expire after a certain amount of time.
        "created_at": time.time(),
        "next_url": next_url,
    }
    request.session[OIDC_STATES_SESSION_KEY] = _clamp_state_entries(states)


def _add_code_verifier_to_session(request, state_key, code_verifier):
    """
    Associate the code verifier with the state.

    This is necessary in order to send it to the token endpoint for verification.
    """
    if (
        OIDC_STATES_SESSION_KEY not in request.session
        or state_key not in request.session[OIDC_STATES_SESSION_KEY]
    ):
        raise ValueError(
            "Missing state in session. State must be added before creating"
            " a PKCE challenge."
        )

    request.session[OIDC_STATES_SESSION_KEY][state_key]["code_verifier"] = code_verifier
    # Django doesn't detect changes to mutable objects stored in the session.
    # Manually mark the session as modified to ensure the changes are saved.
    request.session.modified = True


def get_authentication_request_parameters(
    *, client_id, redirect_uri, state, scope="openid email profile", **extra_params
):
    """
    Prepares the parameters for an authentication request.

    Arguments:
        client_id (str):
            The client ID provided by the OpenID Connect provider.
        redirect_uri (str):
            The absolute URL to redirect the user to after the authentication.
        state (str):
            A unique value to prevent CSRF attacks.
        scope (str):
            The requested scope for the authentication.
        **extra_params:
            Additional parameters to include in the authentication request.

    Returns:
        dict: The parameters for the authentication request.
    """
    # 2.1.1.1. Request Parameters
    # https://openid.net/specs/openid-connect-basic-1_0.html#RequestParameters
    return extra_params | {
        "response_type": "code",
        "client_id": client_id,
        "scope": scope,
        "redirect_uri": redirect_uri,
        "state": state,
    }


def create_pkce_challenge(request, *, state_key):
    """
    Prepares the PKCE challenge for an authentication request.

    Associates the code verifier with the state so it can be sent to
    the token endpoint for verification during the token exchange.
    """
    # 4.1. Client Creates a Code Verifier
    # code_verifier [is a] [...] random STRING with a minimum length
    # of 43 characters and a maximum length of 128 characters.
    # https://www.rfc-editor.org/rfc/rfc7636.html#section-4.1

    # 64 bytes, encoded in base64, is 86 characters long.
    # This is within the recommended range of 43 to 128 characters.
    code_verifier = secrets.token_urlsafe(64)
    _add_code_verifier_to_session(request, state_key, code_verifier)

    # 4.2. Client Creates the Code Challenge
    # "S256" is Mandatory To Implement (MTI) on the server. Clients are
    # permitted to use "plain" only if they cannot support "S256" for some
    # technical reason [...].
    # https://www.rfc-editor.org/rfc/rfc7636.html#section-4.2

    # The code challenge is the SHA-256 hash of the code verifier, encoded
    # as a URL-safe base64 string without padding.
    code_challenge = (
        base64.urlsafe_b64encode(
            hashlib.sha256(code_verifier.encode("ascii")).digest()
        ).decode("ascii")
    ).rstrip("=")  # Strip padding

    # 4.3. Client Sends the Code Challenge with the Authorization Request
    # https://www.rfc-editor.org/rfc/rfc7636.html#section-4.3
    return {
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
    }


def prepare_authentication_request(
    request,
    *,
    client,
    callback_url,
    next_url=None,
    **extra_params,
):
    """
    Prepares an authentication request for an OpenID Connect Authorization Code Flow.

    Arguments:
        request (HttpRequest):
            The current HTTP request.
        client (OIDCClient):
            The OpenID Connect client to use for the authentication request.
        callback_url (str):
            The (relative) URL to redirect the user to after the authentication.
        next_url (str | None):
            The URL to redirect the user to after completing the entire OIDC flow.
        extra_params (dict):
            Additional parameters to include in the authentication request.

    Returns:
        str: The URL to redirect the user to for the authentication.
    """
    # 2.1.1. Client Prepares Authentication Request
    # https://openid.net/specs/openid-connect-basic-1_0.html#AuthenticationRequest
    state_key = secrets.token_urlsafe(32)
    _add_state_to_session(request, state_key, next_url=next_url)

    request_parameters = get_authentication_request_parameters(
        client_id=client.client_id,
        redirect_uri=_build_absolute_uri(request, client, callback_url),
        state=state_key,
        scope=client.scope,
        **extra_params,
    )

    if client.has_pkce_support:
        # Add PKCE parameters to the request.
        request_parameters |= create_pkce_challenge(request, state_key=state_key)

    return urljoin(
        client.authorization_endpoint,
        f"?{urlencode(request_parameters)}",
    )


def _cull_expired_states(states):
    """
    Removes stale state entries from the session.

    The state is considered stale if it has been stored for longer
    than the configured TTL.
    """
    now = time.time()
    return {
        state_key: state
        for state_key, state in states.items()
        if now - state.get("created_at", 0) <= _OIDC_STATE_TTL
    }


def _pop_state_from_session(request, state_key):
    """
    Returns the state stored in the session for the given state ID.

    The state is removed from the session once it has been retrieved.
    If the state is not found, returns None.
    """
    # Get the known states from the session, culling any stale states.
    states = _cull_expired_states(request.session.get(OIDC_STATES_SESSION_KEY, {}))
    # Remove the requested state from the known states.
    state = states.pop(state_key, None)
    # Update the session with the modified states. This is necessary to
    # ensure that the state is not used more than once.
    request.session[OIDC_STATES_SESSION_KEY] = states
    return state


def validate_authentication_callback(request):
    """
    Validates the callback from an OIDC authentication request.

    Arguments:
        request (HttpRequest):
            The current HTTP request.

    Returns:
        tuple (str, dict):
            A tuple containing the code and state associated with the callback.

    Raises:
        OAuth2Error: If the callback contains an error.
        OIDCError: If the callback is invalid.
    """
    # 2.1.5. Authorization Server Sends End-User Back to Client
    # Once the authorization is determined, the Authorization Server
    # returns a successful response or an error response.
    # https://openid.net/specs/openid-connect-basic-1_0.html#CodeResponse

    # 2.1.5.1. End-User Grants Authorization
    # If the End-User grants the access request, the Authorization Server
    # Issues a code and delivers it to the Client [...].
    # https://openid.net/specs/openid-connect-basic-1_0.html#CodeOK
    code = request.GET.get("code")
    state_key = request.GET.get("state")

    # 2.1.5.2. End-User Denies Authorization or Invalid Request
    # If the End-User denies the authorization or the End-User
    # authentication fails, the Authorization Server MUST return the error
    # Authorization Response as defined in Section 4.1.2.1 of OAuth 2.0
    # https://www.rfc-editor.org/rfc/rfc6749.html#section-4.1.2.1
    # https://openid.net/specs/openid-connect-basic-1_0.html#CodeAuthzError
    error = request.GET.get("error")
    if error:
        # Remove the state from the session, authentication failed
        # and the state should not be used again.
        _pop_state_from_session(request, state_key)
        raise OAuth2Error(
            error,
            description=request.GET.get("error_description"),
            uri=request.GET.get("error_uri"),
        )

    for param, value in (("code", code), ("state", state_key)):
        if not value:
            # Remove the state from the session, authentication failed
            # and the state should not be used again.
            _pop_state_from_session(request, state_key)
            raise OIDCError(f"Missing {param!r} in the authentication response.")

    state = _pop_state_from_session(request, state_key)
    if state is None:
        # State is not present in the session (invalid or expired).
        raise InvalidOIDCStateError(
            "Invalid 'state' parameter in the authentication response."
        )

    return code, state


def obtain_tokens(request, *, state, client, code, callback_url):
    """
    Obtains the tokens from an OIDC authentication request.

    Arguments:
        request (HttpRequest):
            The current HTTP request.
        state (dict):
            The state associated with the authentication request.
        client (OIDCClient):
            The OpenID Connect client to use for the authentication request.
        code (str):
            The code received in the callback from the authentication request.
        callback_url (str):
            The (relative) URL to redirect the user to after authentication.

    Returns:
        dict: The token response from the OpenID Connect provider.
    """
    # 2.1.6. Client Obtains ID Token and Access Token
    # https://openid.net/specs/openid-connect-basic-1_0.html#ObtainingTokens

    # 2.1.6.1. Client Sends Code
    # https://openid.net/specs/openid-connect-basic-1_0.html#TokenRequest

    redirect_uri = _build_absolute_uri(request, client, callback_url)
    redirect_uri_origin = "://".join(urlsplit(redirect_uri)[:2])

    token_request_data = {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": redirect_uri,
        "client_id": client.client_id,
    }
    token_request_headers = {
        "Accept": "application/json",
        # Some providers (e.g. Microsoft) require the Origin header
        # to be present and equal the redirect URI origin.
        "Origin": redirect_uri_origin,
    }

    if client.client_secret:
        # Some providers require the client secret to be included in the token request.
        if client.has_basic_auth_support:
            # The Client MUST authenticate to the Token Endpoint using the
            # HTTP Basic method [...].
            basic_auth_credentials = base64.b64encode(
                f"{client.client_id}:{client.client_secret}".encode()
            ).decode()
            token_request_headers |= {
                "Authorization": f"Basic {basic_auth_credentials}"
            }
        else:
            # Some providers (e.g. LinkedIn) require the client secret to be included
            # in the request body instead.
            token_request_data |= {"client_secret": client.client_secret}

    if client.has_pkce_support:
        if "code_verifier" not in state:
            raise OIDCError("Missing 'code_verifier' in state.")
        # 4.5. Client Sends [...] the Code Verifier to the Token Endpoint
        # https://www.rfc-editor.org/rfc/rfc7636.html#section-4.5
        token_request_data["code_verifier"] = state["code_verifier"]

    # 2.1.6.2. Client Receives Tokens
    # https://openid.net/specs/openid-connect-basic-1_0.html#TokenOK
    return requests.post(
        client.token_endpoint,
        data=token_request_data,
        headers=token_request_headers,
        # Timeouts in seconds
        timeout=(
            5,  # Connect timeout
            30,  # Read timeout
        ),
    ).json()


def parse_id_token(raw_id_token, *, client):
    """
    Asserts that the given ID Token is valid and issued by the expected OIDC provider.

    Arguments:
        raw_id_token (str):
            The ID Token to validate.
        client (OIDCClient):
            The OpenID Connect client used to obtain the token.

    Returns:
        dict: The claims from the ID Token.

    Raises:
        OIDCError: If the ID Token is invalid. The authentication process
                   should be aborted and the token should not be used.
    """
    # 2.2. ID Token
    # The ID Token is a security token that contains Claims about the
    # authentication of an End-User by an Authorization Server when using a
    # Client, and potentially other requested Claims. The ID Token is
    # represented as a JSON Web Token (JWT).
    #
    # https://openid.net/specs/openid-connect-basic-1_0.html#IDToken

    # 2.2.1. ID Token Validation
    # The Client MUST validate the ID Token in the Token Response.
    #
    # If any of the validation procedures [...] fail, any operations requiring
    # the information that failed to correctly validate MUST be aborted
    # and the information that failed to validate MUST NOT be used.
    #
    # https://openid.net/specs/openid-connect-basic-1_0.html#IDTokenValidation

    # Parse the token and validate in a way that goes beyond the basic
    # validation described in the OpenID Connect Basic Implementer's Guide.
    # Besides checking the claims, this also validates that the token is
    # signed with the expected algorithm and that the signature is valid.

    # To check the signature, the public key(s) of the OpenID Connect provider
    # must be obtained.
    keys = jwks.get_oidc_client_jwks(client)
    if keys is None:
        raise OIDCError(
            f"Unable to get signing keys for {client.provider_key!r}."
            " The ID Token cannot be validated."
        )

    try:
        # Parse the token, check the signature and perform basic claim validation.
        id_token = jwt.JWT(
            jwt=raw_id_token,
            # Only accept tokens signed with RSA using SHA-256 hash algorithm.
            algs=["RS256"],
            # A signed token is expected (as opposed to an encrypted token).
            expected_type="JWS",
            # Use the public keys of the provider to verify the signature.
            key=keys,
            # Make sure claims are available for further validation.
            check_claims={
                "sub": None,  # Present
                "iss": None,  # Present (value is checked later)
                "aud": client.client_id,  # Must match the client ID
                "exp": None,  # Not expired (current time + 1 minute leeway)
                "iat": None,  # Present
            },
        )
        claims = json.loads(id_token.claims)
    except jwt.JWTMissingKey:
        # The token is signed with an unknown key.
        raise OIDCError(
            f"ID Token from {client.provider_key!r} is signed with an unknown key."
        ) from None
    except (jwt.JWTInvalidClaimValue, jwt.JWTMissingClaim) as exc:
        # Claim verification failed.
        raise OIDCError(
            f"ID Token from {client.provider_key!r} has invalid or missing"
            f" claims: {exc}."
        ) from None
    except jwt.JWTExpired as exc:
        # Token is expired.
        raise OIDCError(
            f"ID Token from {client.provider_key!r} has expired: {exc}."
        ) from None
    except (ValueError, jwt.JWException) as exc:
        # Token is invalid.
        raise OIDCError(f"ID Token from {client.provider_key!r} is invalid.") from exc
    else:
        # Check if the token is currently valid.
        if "nbf" in claims and claims["nbf"] > time.time():
            # The token is not valid yet.
            raise OIDCError(f"ID Token from {client.provider_key!r} is not yet valid.")

        # The token is valid, and not expired.
        # The required claims are available for further validation.

        # Check the issuer
        if claims["iss"] != (expected_issuer := client.get_issuer(claims=claims)):
            # The issuer is not the expected one.
            raise OIDCError(
                f"ID Token from {client.provider_key!r} is not issued by"
                f" {expected_issuer!r}, got {claims['iss']!r}."
            )

        # Make sure the nonce claim is absent or empty. It is not sent in the
        # authentication request, so it should not be present in the ID Token.
        if claims.get("nonce"):
            raise OIDCError(
                f"ID Token from {client.provider_key!r} contains an unexpected"
                f" 'nonce' claim."
            )

        # Everything checks out.
        return claims


def get_user_info(*, client, access_token, claims):
    """
    Obtains the user information from an OpenID Connect provider.

    Arguments:
        client (OIDCClient):
            The OpenID Connect client used to obtain the token.
        access_token (str):
            The access token to use to retrieve the user information.
        claims (dict):
            The claims from the ID Token

    Returns:
        dict: The user information from the OpenID Connect provider.
    """
    # 2.3.1. UserInfo Request
    # Clients send requests to the UserInfo Endpoint to obtain Claims about
    # the End-User using an Access Token obtained through [...] Authentication.
    # https://openid.net/specs/openid-connect-basic-1_0.html#UserInfoRequest

    try:
        # The request SHOULD use the HTTP GET method [...]
        response = requests.get(
            client.userinfo_endpoint,
            headers={
                # the Access Token SHOULD be sent using the Authorization header field.
                "Authorization": f"Bearer {access_token}",
                "Accept": "application/json",
            },
            # Generous timeouts
            timeout=(
                5,  # Connect timeout
                30,  # Read timeout
            ),
        )
    except requests.RequestException as exc:
        raise OIDCError(
            f"Failed to fetch user information from {client.provider_key!r}"
            f" from {client.userinfo_endpoint!r}."
        ) from exc

    try:
        response.raise_for_status()
    except requests.RequestException as exc:
        raise OIDCError(
            f"Error after fetching user information from {client.provider_key!r}"
            f" from {client.userinfo_endpoint!r}: {response.status_code}."
        ) from exc

    # 2.3.2. Successful UserInfo Response
    # https://openid.net/specs/openid-connect-basic-1_0.html#UserInfoResponse
    try:
        # The UserInfo Claims MUST be returned as the members of a JSON object.
        user_info = client.normalize_userinfo(userinfo=response.json())
    except json.JSONDecodeError as exc:
        raise OIDCError(
            f"Failed to parse user information from {client.provider_key!r}"
            f" from {client.userinfo_endpoint!r}."
        ) from exc

    # The sub Claim in the UserInfo Response MUST be verified to exactly match
    # the sub Claim in the ID Token; if they do not match, the UserInfo Response
    # values MUST NOT be used.
    if "sub" not in user_info or user_info["sub"] != claims["sub"]:
        raise OIDCError(
            f"User information from {client.provider_key!r} does not match the"
            f" ID token 'sub' claim."
        )

    return user_info


_AuthenticationResult = collections.namedtuple(
    "_AuthenticationResult", ["tokens", "claims", "user_info", "next_url"]
)


def handle_authentication_callback(request, *, client, callback_url):
    """
    Handles the callback from an OIDC authentication request.

    Arguments:
        request (HttpRequest):
            The current HTTP request.
        client (OIDCClient):
            The OpenID Connect client to use for the authentication request.
        callback_url (str):
            The (relative) URL to redirect the user to after authentication.

    Returns:
        _AuthenticationResult:
            A named tuple with the tokens, claims, and user
            information obtained from the authentication

    Raises:
        OAuth2Error: If the callback contains an error.
        OIDCError: If the callback is invalid.
    """
    code, state = validate_authentication_callback(request)
    token_response = obtain_tokens(
        request,
        state=state,
        client=client,
        code=code,
        callback_url=callback_url,
    )
    if "error" in token_response:
        raise OAuth2Error(
            token_response["error"],
            description=token_response.get("error_description"),
            uri=token_response.get("error_uri"),
        )
    claims = parse_id_token(token_response.get("id_token"), client=client)
    user_info = get_user_info(
        client=client, access_token=token_response["access_token"], claims=claims
    )
    return _AuthenticationResult(
        tokens=token_response,
        claims=claims,
        user_info=user_info,
        next_url=state.get("next_url"),
    )



from oauth2_provider.contrib.rest_framework import OAuth2Authentication
from rest_framework import mixins, viewsets
from rest_framework.authentication import SessionAuthentication
from rest_framework.permissions import IsAuthenticated

from django.contrib.auth import get_user_model
from django.http import Http404

from .serializers import UserSerializer

UserModel = get_user_model()


class UserViewSet(
    mixins.RetrieveModelMixin, mixins.UpdateModelMixin, viewsets.GenericViewSet
):
    authentication_classes = [SessionAuthentication, OAuth2Authentication]
    serializer_class = UserSerializer
    permission_classes = [IsAuthenticated]
    queryset = UserModel.objects.all()

    def get_object(self):
        # Users can only ever access themselves using the "me" shortcut.
        if self.kwargs.get(self.lookup_url_kwarg or self.lookup_field) == "me":
            return self.request.user
        raise Http404



from http import HTTPStatus
from unittest import mock

from django.test import TestCase, override_settings
from django.urls import reverse

from hidp.accounts import auth as hidp_auth
from hidp.test.factories import user_factories


@override_settings(LOGIN_REDIRECT_URL="/", LANGUAGE_CODE="en")
class TestLogout(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.logout_url = reverse("hidp_accounts:logout")

    def test_get_logout(self):
        """GET is not allowed for the logout view."""
        response = self.client.get(self.logout_url)
        self.assertEqual(HTTPStatus.METHOD_NOT_ALLOWED, response.status_code)

    @mock.patch("hidp.accounts.views.hidp_auth.logout", wraps=hidp_auth.logout)
    def test_post_logout(self, mock_logout):
        self.client.post(self.logout_url)
        mock_logout.assert_called_once()

    def test_post_logout_no_login(self):
        """POST is allowed for the logout view, even if the user is not logged in."""
        session_key = self.client.session.session_key
        response = self.client.post(self.logout_url)
        self.assertRedirects(response, "/", fetch_redirect_response=False)
        self.assertNotEqual(session_key, self.client.session.session_key)

    def test_post_logout_with_login(self):
        """POST is allowed for the logout view."""
        self.client.force_login(self.user)
        session_key = self.client.session.session_key
        response = self.client.post(self.logout_url)
        self.assertRedirects(response, "/", fetch_redirect_response=False)
        self.assertNotEqual(session_key, self.client.session.session_key)

    def test_post_logout_with_safe_next_param(self):
        """Redirects to the next url, if it's safe."""
        response = self.client.post(
            self.logout_url,
            {"next": "/example/"},
        )
        self.assertRedirects(response, "/example/", fetch_redirect_response=False)

    def test_post_logout_with_unsafe_next_param(self):
        """Redirects to the default url, if the next url is unsafe."""
        response = self.client.post(
            self.logout_url,
            {"next": "https://example.com/"},
        )
        self.assertRedirects(response, "/", fetch_redirect_response=False)



"""Microsoft OpenID Connect provider."""

from . import base

# Follow the instructions from Microsoft on how to set up OpenID Connect on the
# Microsoft Identity Platform:
#
# https://learn.microsoft.com/en-us/entra/identity-platform/v2-protocols-oidc
#
# Quick start:
#
# 1. Register a new application in the Entra portal:
#    https://entra.microsoft.com/#view/Microsoft_AAD_RegisteredApps/ApplicationsListBlade/quickStartType~/null/sourceType/Microsoft_AAD_IAM
#
#    - Set the application type to "Single Page Application".
#      This enables the Code Flow with PKCE, and avoids the need for a client secret.
#    - Pick the broadest account type possible (organization, personal, etc.)
#    - Set the redirect URI to: https://<domain>/login/oidc/callback/microsoft/
#      Private domains (e.g. *.local, *.test) are allowed.


class MicrosoftOIDCClient(base.OIDCClient):
    provider_key = "microsoft"

    # Microsoft OpenID Connect configuration:
    # https://login.microsoftonline.com/common/v2.0/.well-known/openid-configuration
    issuer = "https://login.microsoftonline.com/{tenantid}/v2.0"
    authorization_endpoint = "https://login.microsoftonline.com/common/oauth2/v2.0/authorize"  # fmt: skip # noqa: E501
    token_endpoint = "https://login.microsoftonline.com/common/oauth2/v2.0/token"  # noqa: S105 (not a secret)
    userinfo_endpoint = "https://graph.microsoft.com/oidc/userinfo"
    jwks_uri = "https://login.microsoftonline.com/common/discovery/v2.0/keys"

    def get_issuer(self, *, claims):
        # Use the tenant ID from the claims to format the issuer URL.
        # The common issuer URL is used for all tenants, but the tenant ID
        # is required for the issuer URL to be valid.
        #
        # This is somewhat strange, but it's how Microsoft has set up
        # their OpenID Connect configuration.
        return (
            self.issuer.format(tenantid=tid)
            if (tid := claims.get("tid"))
            else self.issuer
        )



import fnmatch
import os

from django.core.management.commands import makemessages

POT_TEMPLATE = r"""
#, fuzzy
msgid ""
msgstr ""
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
""".strip()


class Command(makemessages.Command):
    """Wraps Django's makemessages command to create translation files for HIdP."""

    msgmerge_options = [
        *makemessages.Command.msgmerge_options,
        "--sort-by-file",
    ]
    msguniq_options = [
        *makemessages.Command.msguniq_options,
        "--sort-by-file",
    ]
    msgattrib_options = [
        *makemessages.Command.msgattrib_options,
        "--sort-by-file",
    ]
    xgettext_options = [
        *makemessages.Command.xgettext_options,
        "--sort-by-file",
    ]

    def add_arguments(self, parser):
        super().add_arguments(parser)
        # Override default options to avoid having to specify them
        # on the command line each time.
        parser.set_defaults(
            # Process all available locales
            all=True,
            # Add location comments, but omit line numbers
            add_location="file",
            # Do not keep obsolete messages
            no_obsolete=True,
            # Do not wrap long messages
            no_wrap=True,
            # Do not remove the .pot file
            keep_pot=True,
            # Additional ignore patterns
            ignore_patterns=[
                # Ignore editable install artifact
                "*.egg-info/*",
                # Ignore docs and tests
                "docs/*",
                "tests/*",
            ],
        )

    def find_files(self, root):
        # The additional ignore patters are not enough to exclude all
        # unwanted files, things like Makefile, README are still found.
        # Instead of adding more ignore patterns, it is easier to just
        # remove any file that is not nested in the hidp package directory.
        return [
            file
            for file in super().find_files(root)
            if fnmatch.fnmatchcase(file.path, f"{root}/hidp/*")
        ]

    def remove_potfiles(self):
        # Even though keep_pot is set to True, the default implementation
        # still removes the .pot file at the start of the process.
        # We do not want this, and instead use this to recreate the .pot
        # file using our own customized template to remove some of
        # the default (and in our opinion, pointless) comments and headers.
        for path in self.locale_paths:
            pot_path = os.path.join(path, f"{self.domain}.pot")
            if not os.path.exists(pot_path):
                continue
            with open(pot_path, "w", encoding="utf-8") as f:
                f.write(POT_TEMPLATE)



import string

from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from hidp.accounts.password_validation import (
    DigitValidator,
    LowercaseValidator,
    SymbolValidator,
    UppercaseValidator,
)


class TestPasswordValidation(SimpleTestCase):
    def test_digit_validator(self):
        validator = DigitValidator()

        self.assertEqual(
            validator.get_help_text(),
            "Your password must contain at least one digit (0-9).",
        )

        with self.subTest("Valid password"):
            validator.validate("P@ssw0rd!")

        with (
            self.subTest("Invalid password"),
            self.assertRaises(ValidationError) as cm,
        ):
            validator.validate("Password!")
        self.assertEqual(cm.exception.code, "password_no_digit")
        self.assertEqual(
            cm.exception.message, "This password does not contain any digits (0-9)."
        )

    def test_uppercase_validator(self):
        validator = UppercaseValidator()

        self.assertEqual(
            validator.get_help_text(),
            "Your password must contain at least one uppercase character (A-Z).",
        )

        with self.subTest("Valid password"):
            validator.validate("P@ssw0rd!")

        with (
            self.subTest("Invalid password"),
            self.assertRaises(ValidationError) as cm,
        ):
            validator.validate("p@ssw0rd!")
        self.assertEqual(cm.exception.code, "password_no_upper")
        self.assertEqual(
            cm.exception.message,
            "This password does not contain any uppercase characters (A-Z).",
        )

    def test_lowercase_validator(self):
        validator = LowercaseValidator()

        self.assertEqual(
            validator.get_help_text(),
            "Your password must contain at least one lowercase character (a-z).",
        )

        with self.subTest("Valid password"):
            validator.validate("P@ssw0rd!")

        with (
            self.subTest("Invalid password"),
            self.assertRaises(ValidationError) as cm,
        ):
            validator.validate("P@SSW0RD!")
        self.assertEqual(cm.exception.code, "password_no_lower")
        self.assertEqual(
            cm.exception.message,
            "This password does not contain any lowercase characters (a-z).",
        )

    def test_symbol_validator(self):
        validator = SymbolValidator()
        punctuation = string.punctuation

        self.assertEqual(
            validator.get_help_text(),
            (
                f"Your password must contain at least one"
                f" special character ({punctuation})."
            ),
        )

        with self.subTest("Valid password"):
            validator.validate("P@ssw0rd!")

        with (
            self.subTest("Invalid password"),
            self.assertRaises(ValidationError) as cm,
        ):
            validator.validate("Password1")
        self.assertEqual(cm.exception.code, "password_no_symbol")
        self.assertEqual(
            cm.exception.message,
            f"This password does not contain any special characters ({punctuation}).",
        )



from django_ratelimit.decorators import ratelimit

_default_rate_limits = [
    ratelimit(key="ip", method=ratelimit.UNSAFE, rate="10/s"),
    ratelimit(key="ip", method=ratelimit.UNSAFE, rate="30/m"),
]


def _apply_rate_limits(*rate_limits, view):
    for rate_limit in rate_limits:
        view = rate_limit(view)
    return view


def rate_limit_default(view):
    return _apply_rate_limits(*_default_rate_limits, view=view)


def rate_limit_strict(view):
    return _apply_rate_limits(
        *_default_rate_limits,
        ratelimit(key="ip", method=ratelimit.ALL, rate="100/15m"),
        view=view,
    )



from django.conf import settings
from django.contrib.auth import models as auth_models
from django.db import models
from django.db.models import functions
from django.utils.translation import gettext_lazy as _

from ..compat.uuid7 import uuid7


class UserQuerySet(models.QuerySet):
    def email_verified(self):
        """
        Only include users that have verified their email address.

        Returns:
            ``QuerySet``: Users that have verified their email address.
        """
        return self.exclude(email_verified__isnull=True)

    def email_unverified(self):
        """
        Only include users that have not verified their email address.

        Returns:
            ``QuerySet``: Users that have not verified their email address.
        """
        return self.filter(email_verified__isnull=True)


class UserManager(auth_models.UserManager.from_queryset(UserQuerySet)):
    """Custom user manager that uses email as the username field."""

    use_in_migrations = True

    def _create_user(self, username, email, password, **extra_fields):
        if not email:
            raise ValueError("User must have an email address")
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """
        Create a new user with the given email and password.

        Prefer using this method over instantiating the user model directly,
        as it ensures that the email address is normalized and the password is hashed.

        Automatically sets ``is_staff`` to ``False`` and ``is_superuser`` to ``False``,
        unless explicitly set otherwise in ``extra_fields``.

        Args:
            email (``str``): The email address of the user.
            password (``str``, `optional`): The password of the user.
            **extra_fields: Additional fields to set on the user.

        Returns:
            ``User``: The newly created user.
        """
        return super().create_user(
            username=email,
            email=email,
            password=password,
            **extra_fields,
        )

    def create_superuser(self, email, password=None, **extra_fields):
        """
        Create a new superuser with the given email and password.

        Automatically sets ``is_staff`` and ``is_superuser`` to ``True``,
        unless explicitly set otherwise in ``extra_fields``.

        Args:
            email (``str``): The email address of the user.
            password (``str``, `optional`): The password of the user.
            **extra_fields: Additional fields to set on the user.

        Returns:
            ``User``: The newly created superuser.
        """
        return super().create_superuser(
            username=email,
            email=email,
            password=password,
            **extra_fields,
        )


class BaseUser(auth_models.AbstractUser):
    """
    Abstract base class that extends Django's default user model.

    Attributes:
        password (``CharField``):
            Hashed password.

        first_name (``CharField``):
            Given name, `optional`

        last_name (``CharField``):
            Family name, `optional`

        is_active (``BooleanField``):
            Whether the user is active (allowed to log in).

            Defaults to ``True``.

        is_staff (``BooleanField``):
            Whether the user is a staff member (allowed to log into the admin site).

            This field is used exclusively by Django's permissions system and
            has no special meaning in HIdP. It is recommended to avoid direct use
            of this field in your application code. Instead, rely on the permissions
            features provided by Django.

            Defaults to ``False``.

        is_superuser (``BooleanField``):
            Whether the user is a superuser (allowed to do anything).

            This field has no additional meaning in HIdP. It is recommended to avoid
            direct use of this field in your application code. Instead, rely on Django's
            group features to assign specific permissions groups of users.

            Defaults to ``False``.

        groups (``ManyToManyField``):
            Groups the user belongs to

        user_permissions (``ManyToManyField``):
            Permissions the user has

        date_joined (``DateTimeField``):
            Date and time when the user was created.

            Defaults to the current date and time.

        last_login (``DateTimeField``):
            Date and time when the user last logged in
            Populated by Django when the ``django.contrib.auth.user_logged_in`` signal
            is sent.

            Defaults to ``None``.

    .. rubric:: Alters the default Django user model with the following modifications
        :heading-level: 3

    Attributes:
        id (``UUIDField``):
            Primary key, a version 7 UUID.

            Altered:
                Was ``int`` in the default Django user model.

        email (``EmailField``):
            Email address, case-insensitive, unique and required.

            Altered:
                Was case-sensitive, not unique and optional in the default Django
                user model.

        username (``None``):
            Altered:
                The username field is removed in favor of the email field

    .. rubric:: Adds the following attributes
        :heading-level: 3

    Attributes:
        email_verified (``DateTimeField``):
            Date and time when the email address was verified.

            Defaults to ``None``.

        agreed_to_tos (``DateTimeField``):
            Date and time when the user agreed to the terms of service.

            Defaults to ``None``

        last_modified (``DateTimeField``):
            Date and time when the user was last modified. Populated by Django when
            the user is saved.

            Defaults to the current date and time.
    """

    # Change the primary key to UUID
    id = models.UUIDField(primary_key=True, default=uuid7, editable=False)
    # Remove the username field
    username = None
    email = models.EmailField(_("email address"), unique=True)
    # Store the date when the email was verified
    email_verified = models.DateTimeField(
        _("email verified"), blank=True, null=True, editable=False
    )
    # Store the last modification date
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)
    # Store when the user agreed to the terms of service
    agreed_to_tos = models.DateTimeField(
        _("agreed to terms of service"),
        blank=True,
        null=True,
    )

    # Use the email field as the username field
    USERNAME_FIELD = "email"
    # Add names as required fields
    REQUIRED_FIELDS = ["first_name", "last_name"]

    objects = UserManager()

    class Meta:
        abstract = True
        verbose_name = _("user")
        verbose_name_plural = _("users")
        constraints = [
            models.UniqueConstraint(
                functions.Lower("email"),
                name="unique_lower_email",
            )
        ]

    @property
    def is_anonymous(self):
        """
        Helper property to find out if user is anonymous or authenticated.

        ``bool``: Always ``False``. As opposed to always ``True`` for ``AnonymousUser``.
        """
        return super().is_anonymous

    @property
    def is_authenticated(self):
        """
        Helper property to find out if user is anonymous or authenticated.

        ``bool``: Always ``True``. As opposed to always ``False`` for ``AnonymousUser``.
        """
        return super().is_authenticated

    def check_password(self, raw_password):
        """
        Check the raw password against the user's hashed password.

        When the password is correct, but uses an outdated hashing algorithm,
        the password is upgraded to use the latest algorithm.

        Will save the user if the password is upgraded.

        Args:
            raw_password (``str``): The raw password to be checked.

        Returns:
            ``True`` if the password is correct, ``False`` otherwise.

        """
        return super().check_password(raw_password)

    def clean(self):
        """
        Normalize the email address by lower-casing the domain part.

        Automatically called before saving the user.
        """
        super().clean()

    def email_user(self, subject, message, from_email=None, **kwargs):  # noqa: D417
        """
        Email this user with the given subject and message.

        If ``from_email`` is not specified ``settings.DEFAULT_FROM_EMAIL`` is used.

        Additional keyword arguments are passed to the `send_mail` function as-is.

        Args:
            subject (``str``): The subject of the email.
            message (``str``): The message of the email.
            from_email (``str``, `optional`): The sender's email address.
        """
        super().email_user(subject, message, from_email=from_email, **kwargs)

    def get_full_name(self):
        """
        Return the first name and the last name, separated by a space.

        Returns:
            ``str``: The full name of the user.
        """
        return super().get_full_name()

    def get_short_name(self):
        """
        Return the first name.

        Returns:
            ``str``: The first name of the user.
        """
        return super().get_short_name()

    def has_usable_password(self):
        """
        Check if the user has a usable password.

        Returns:
            ``bool``: ``True`` if the user has a password set and it doesn't begin with
            the unusable password prefix.
        """
        return super().has_usable_password()

    def save(self, *args, update_fields=None, **kwargs):
        """
        Save the user to the database.

        Altered:
            Always normalizes the email address before saving.
        """
        self.clean()  # Always normalize the email address
        super().save(*args, update_fields=update_fields, **kwargs)

    def set_password(self, raw_password):
        """
        Set the user's password field to the hashed value of the raw password.

        The user is **not** saved after setting the password.

        Args:
            raw_password (``str``): The raw password to be hashed.
        """
        super().set_password(raw_password)

    def set_unusable_password(self):
        """Set the user's password field to a value that will never be a valid hash."""
        super().set_unusable_password()


class EmailChangeRequest(models.Model):
    """
    A user's request to change their email and it's confirmation state.

    Attributes:
        uuid (``UUIDField``):
            Primary key, a version 7 UUID.

        user (``OneToOneField``):
            The user that requested the email change.

        proposed_email (``EmailField``):
            The email address the user wants to change to.

        confirmed_by_proposed_email (``BooleanField``):
            Whether the email change has been confirmed by the proposed email address.

            Defaults to ``False``.

        confirmed_by_current_email (``BooleanField``):
            Whether the email change has been confirmed by the current email address.

            Defaults to ``False``.

        created (``DateTimeField``):
            Date and time when the email change request was created.

        modified (``DateTimeField``):
            Date and time when the email change request was last modified.
    """

    id = models.UUIDField(default=uuid7, editable=False, primary_key=True)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="email_change_request",
        verbose_name=_("user"),
    )
    current_email = models.EmailField(_("current email address"))
    proposed_email = models.EmailField(_("proposed email address"))
    confirmed_by_proposed_email = models.BooleanField(
        _("confirmed by proposed email"),
        default=False,
    )
    confirmed_by_current_email = models.BooleanField(
        _("confirmed by current email"),
        default=False,
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    last_modified = models.DateTimeField(_("last modified"), auto_now=True)

    class Meta:
        verbose_name = _("email change request")
        verbose_name_plural = _("email change requests")

    def __str__(self):
        return f"Email change request for {self.current_email} to {self.proposed_email}"

    def is_complete(self):
        """
        Check if the email change request is complete.

        Returns:
            ``bool``: ``True`` if the request is complete, ``False`` otherwise.
        """
        return self.confirmed_by_proposed_email and self.confirmed_by_current_email



from django.test import SimpleTestCase

from hidp.config import configure_oidc_clients
from hidp.config.oidc_clients import (
    get_oidc_client,
    get_oidc_client_or_none,
    get_registered_oidc_clients,
)

from ..test_federated.test_providers.example import ExampleOIDCClient


class ExampleOIDCClient2(ExampleOIDCClient):
    # Same as the ExampleOIDCClient, but with a different provider key.
    provider_key = "example-2"


class TestOIDCClientsRegistry(SimpleTestCase):
    def setUp(self):
        # Empty the registry before each test, by configuring zero clients.
        configure_oidc_clients()

    def test_register_client(self):
        """Registering a client makes it retrievable."""
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))
        client = get_oidc_client("example")
        self.assertIsInstance(client, ExampleOIDCClient)

    def test_register_duplicate_client(self):
        """Registering a client with a duplicate provider key raises an error."""
        with self.assertRaisesMessage(
            ValueError,
            "Duplicate provider key: 'example'",
        ):
            configure_oidc_clients(
                ExampleOIDCClient(client_id="test"),
                ExampleOIDCClient(client_id="test-2"),
            )

    def test_register_multiple_clients(self):
        """Registering multiple clients makes them all retrievable."""
        configure_oidc_clients(
            ExampleOIDCClient(client_id="test"),
            ExampleOIDCClient2(client_id="test-2"),
        )
        client = get_oidc_client("example")
        self.assertIsInstance(client, ExampleOIDCClient)
        client = get_oidc_client("example-2")
        self.assertIsInstance(client, ExampleOIDCClient2)

    def test_register_invalid_client(self):
        """Registering anything other than an OIDCClient raises a TypeError."""
        with self.assertRaisesMessage(
            TypeError,
            "Expected OIDCClient, got 'str'",
        ):
            configure_oidc_clients("not a client")

    def test_register_twice(self):
        """Registering clients twice overwrites the previous registration."""
        configure_oidc_clients(
            ExampleOIDCClient(client_id="test"),
            ExampleOIDCClient2(client_id="test-2"),
        )
        configure_oidc_clients(ExampleOIDCClient(client_id="test-2"))
        client = get_oidc_client("example")
        # The second registration overwrites the first one
        self.assertIsInstance(client, ExampleOIDCClient)
        self.assertEqual(client.client_id, "test-2")

    def test_get_unregistered_client(self):
        """Retrieving an unregistered client raises a KeyError."""
        with self.assertRaisesMessage(
            KeyError,
            "No OIDC client registered for provider key: 'example'",
        ):
            get_oidc_client("example")

    def test_get_registered_oidc_clients(self):
        """Retrieving all registered clients returns a list in registration order."""
        clients = [
            ExampleOIDCClient(client_id="test"),
            ExampleOIDCClient2(client_id="test-2"),
        ]
        configure_oidc_clients(*clients)
        registered_clients = get_registered_oidc_clients()
        self.assertEqual(len(clients), len(registered_clients))
        self.assertEqual(clients, registered_clients)

        with self.subTest("Mutation"):
            # Mutating the returned list should not affect the registry.
            registered_clients[0] = None
            self.assertEqual(clients, get_registered_oidc_clients())

    def test_get_oidc_client_or_none(self):
        """Retrieving a client or None should return the client or None."""
        configure_oidc_clients(ExampleOIDCClient(client_id="test"))

        with self.subTest("Registered client"):
            client = get_oidc_client_or_none("example")
            self.assertIsInstance(client, ExampleOIDCClient)

        with self.subTest("Unknown client"):
            client = get_oidc_client_or_none("non-existent")
            self.assertIsNone(client)

        with self.subTest("None key"):
            client = get_oidc_client_or_none(None)
            self.assertIsNone(client)



import django.db.models.deletion

from django.conf import settings
from django.db import migrations, models

import hidp.compat.uuid7


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="OpenIdConnection",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=hidp.compat.uuid7.uuid7,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="openid_connections",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("created_at", models.DateTimeField("created at", auto_now_add=True)),
                ("last_modified", models.DateTimeField("last modified", auto_now=True)),
                ("provider_key", models.CharField(max_length=100)),
                ("issuer_claim", models.CharField(max_length=255)),
                ("subject_claim", models.CharField(max_length=255)),
            ],
            options={
                "verbose_name": "OpenID connection",
                "verbose_name_plural": "OpenID connections",
                "unique_together": {("provider_key", "issuer_claim", "subject_claim")},
            },
        ),
    ]



import contextlib

from django.conf import settings
from django.http import HttpResponseRedirect
from django.utils import translation


def _get_supported_language_variant(language_code):
    with contextlib.suppress(LookupError):
        return translation.get_supported_language_variant(language_code)


def _get_language_from_ui_locales(request):
    ui_locales = request.GET.get("ui_locales", "").strip()
    for ui_locale in ui_locales.split():
        if not translation.check_for_language(ui_locale):
            continue
        if lang_code := _get_supported_language_variant(ui_locale):
            return lang_code
    return None


class UiLocalesMiddleware:
    """
    Set the language preference based on the 'ui_locales' parameter in the query.

    The 'ui_locales' parameter is a space-separated list of language tags, ordered by
    preference. The first language tag that is supported by the application is stored
    in a cookie. If no supported language is found, the language cookie is not set.

    This cookie is used by Django's LocaleMiddleware to set the language for
    the request and may also be set using Django's set_language view.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if lang_code := _get_language_from_ui_locales(request):
            # The 'ui_locales' parameter contains a supported language.
            # Remove the parameter from the query string:
            query = request.GET.copy()
            query.pop("ui_locales")
            request.META["QUERY_STRING"] = query.urlencode()
            response = HttpResponseRedirect(request.build_absolute_uri())
            # Set the language preference cookie,
            # exactly as Django's set_language view does:
            response.set_cookie(
                settings.LANGUAGE_COOKIE_NAME,
                lang_code,
                max_age=settings.LANGUAGE_COOKIE_AGE,
                path=settings.LANGUAGE_COOKIE_PATH,
                domain=settings.LANGUAGE_COOKIE_DOMAIN,
                secure=settings.LANGUAGE_COOKIE_SECURE,
                httponly=settings.LANGUAGE_COOKIE_HTTPONLY,
                samesite=settings.LANGUAGE_COOKIE_SAMESITE,
            )
            return response
        return self.get_response(request)



from unittest import mock

from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse
from django.test import RequestFactory, TestCase

from hidp.otp.decorators import otp_exempt
from hidp.otp.middleware import (
    OTPRequiredMiddleware,
    OTPSetupRequiredIfStaffUserMiddleware,
    OTPVerificationRequiredIfConfiguredMiddleware,
)
from hidp.test.factories import otp_factories, user_factories


class OTPMiddlewareTestBase(TestCase):
    def setUp(self):
        self.request_factory = RequestFactory()

    @staticmethod
    def verify_user(user, *, verified):
        user.is_verified = lambda: verified

    def assertMiddlewareRedirects(self, response, expected_url, message=None):  # noqa: N802
        """
        Replacement for self.assertRedirects that works without the test client.

        The result of a middleware process_view call is a redirect response, but we
        cannot use the test client to check the response. This method checks the
        response manually. Additionally, the result of process_view may be None if
        the middleware does not redirect, so we need to check for that as well.
        """
        self.assertIsNotNone(response, "Expected a redirect response.")
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, expected_url, message)


def basic_view(request):
    return HttpResponse("Hello, world!")


@otp_exempt
def exempt_view(request):
    return HttpResponse("I'm exempt from OTP verification.")


class TestOTPRequiredMiddleware(OTPMiddlewareTestBase):
    @classmethod
    def setUpTestData(cls):
        cls.middleware = OTPRequiredMiddleware(lambda request: None)
        cls.user = user_factories.VerifiedUserFactory()

    def setUp(self):
        super().setUp()
        self.verify_user(self.user, verified=False)

    def test_anonymous_user(self):
        """Anonymous users should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = AnonymousUser()

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected anonymous users to not need to verify OTP.",
        )

    def test_unverified_user(self):
        """Unverified users should need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.user

        self.assertTrue(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected unverified users to need to verify OTP.",
        )

    def test_verified_user(self):
        """Verified users should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.user
        self.verify_user(request.user, verified=True)

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected verified users to not need to verify OTP.",
        )

    def test_exempt_view(self):
        """Views marked as exempt should not require verification."""
        request = self.request_factory.get("/")
        request.user = self.user

        self.assertFalse(
            self.middleware.request_needs_verification(request, exempt_view),
            msg="Expected exempt views to not require OTP verification.",
        )

    def test_process_view_redirects_unconfirmed_to_setup(self):
        """The middleware should redirect unconfirmed users to the OTP setup view."""
        request = self.request_factory.get("/some-path/")
        request.user = self.user

        response = self.middleware.process_view(request, basic_view, [], {})
        self.assertMiddlewareRedirects(
            response,
            "/manage/otp/setup/?next=%2Fsome-path%2F",
            "Expected redirect to OTP setup view.",
        )

    def test_process_view_redirects_confirmed_to_verify(self):
        """The middleware should redirect confirmed users to the verification view."""
        request = self.request_factory.get("/some-path/")
        request.user = self.user
        otp_factories.TOTPDeviceFactory(user=request.user, confirmed=True)

        response = self.middleware.process_view(request, basic_view, [], {})
        self.assertMiddlewareRedirects(
            response,
            "/otp/verify/?next=%2Fsome-path%2F",
            "Expected redirect to OTP verification view.",
        )

    def test_process_view_does_not_redirect(self):
        """The middleware should not redirect if verification is not required."""
        request = self.request_factory.get("/")
        request.user = AnonymousUser()

        response = self.middleware.process_view(request, basic_view, [], {})

        self.assertIsNone(response)


class TestOTPRequiredIfConfiguredMiddleware(OTPMiddlewareTestBase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.VerifiedUserFactory()
        cls.confirmed_user = user_factories.VerifiedUserFactory()
        otp_factories.TOTPDeviceFactory(user=cls.confirmed_user, confirmed=True)
        otp_factories.StaticDeviceFactory(user=cls.confirmed_user, confirmed=True)

        cls.middleware = OTPVerificationRequiredIfConfiguredMiddleware(
            lambda request: None
        )

    def setUp(self):
        super().setUp()

        # The middleware expects the OTPMiddleware to have already run, so we need to
        # manually set the user's verified property.
        self.verify_user(self.user, verified=False)
        self.verify_user(self.confirmed_user, verified=False)

    @mock.patch("hidp.otp.middleware.user_has_device")
    def test_anonymous_user(self, mock_user_has_device):
        """Anonymous users should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = AnonymousUser()

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected anonymous users to not need to verify OTP.",
        )
        mock_user_has_device.assert_not_called()

    def test_user_without_otp(self):
        """Users without OTP devices should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.user

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected users without OTP devices to not need to verify OTP.",
        )

    def test_user_with_unconfirmed_otp(self):
        """Users with unconfirmed OTP devices should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.user
        otp_factories.TOTPDeviceFactory(user=self.user, confirmed=False)

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected users with unconfirmed OTP devices to not need to verify"
            " OTP.",
        )

    def test_user_with_confirmed_otp(self):
        """Users with confirmed OTP devices should need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.confirmed_user

        self.assertTrue(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected users with confirmed OTP devices to need to verify OTP.",
        )

    @mock.patch("hidp.otp.middleware.user_has_device")
    def test_verified_user(self, mock_user_has_device):
        """Users who have already verified their OTP should not need to verify again."""
        request = self.request_factory.get("/")
        request.user = self.confirmed_user
        self.verify_user(request.user, verified=True)

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected users who have already verified their OTP to not need to "
            "verify again.",
        )
        mock_user_has_device.assert_not_called()

    @mock.patch("hidp.otp.middleware.user_has_device")
    def test_exempt_view(self, mock_user_has_device):
        """Views marked as exempt should not require verification."""
        request = self.request_factory.get("/")
        request.user = self.confirmed_user

        self.assertFalse(
            self.middleware.request_needs_verification(request, exempt_view),
            msg="Expected exempt views to not require OTP verification.",
        )
        mock_user_has_device.assert_not_called()

    def test_process_view_redirects(self):
        """The middleware should redirect to the OTP verification view if required."""
        request = self.request_factory.get("/some-path/")
        request.user = self.confirmed_user

        response = self.middleware.process_view(request, basic_view, [], {})
        self.assertMiddlewareRedirects(
            response,
            "/otp/verify/?next=%2Fsome-path%2F",
            "Expected redirect to OTP verification view.",
        )

    def test_process_view_does_not_redirect(self):
        """The middleware should not redirect if verification is not required."""
        request = self.request_factory.get("/")
        request.user = AnonymousUser()

        response = self.middleware.process_view(request, basic_view, [], {})

        self.assertIsNone(response)


class TestOTPRequiredIfStaffUser(OTPMiddlewareTestBase):
    @classmethod
    def setUpTestData(cls):
        cls.middleware = OTPSetupRequiredIfStaffUserMiddleware(lambda request: None)
        cls.unconfirmed_staff_user = user_factories.VerifiedUserFactory(is_staff=True)
        cls.verify_user(cls.unconfirmed_staff_user, verified=False)

        cls.confirmed_staff_user = user_factories.VerifiedUserFactory(is_staff=True)
        otp_factories.TOTPDeviceFactory(user=cls.confirmed_staff_user, confirmed=True)
        cls.verify_user(cls.confirmed_staff_user, verified=False)

    def test_anonymous_user(self):
        """Anonymous users should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = AnonymousUser()

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected anonymous users to not need to verify OTP.",
        )

    def test_non_staff_user(self):
        """Non-staff users should not need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = user_factories.VerifiedUserFactory(is_staff=False)
        self.verify_user(request.user, verified=False)

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected non-staff users to not need to verify OTP.",
        )

    def test_staff_user(self):
        """Staff users should need to verify OTP."""
        request = self.request_factory.get("/")
        request.user = self.unconfirmed_staff_user

        self.assertTrue(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected staff users to need to verify OTP.",
        )

    def test_verified_staff_user(self):
        """Staff users who have already verified their OTP should not need to verify."""
        request = self.request_factory.get("/")
        request.user = self.confirmed_staff_user
        self.verify_user(request.user, verified=True)

        self.assertFalse(
            self.middleware.request_needs_verification(request, basic_view),
            msg="Expected staff users who have already verified their OTP to not need "
            "to verify again.",
        )

    def test_staff_user_exempt_view(self):
        """Staff users should not need to verify OTP for exempt views."""
        request = self.request_factory.get("/")
        request.user = self.unconfirmed_staff_user

        self.assertFalse(
            self.middleware.request_needs_verification(request, exempt_view),
            msg="Expected staff users to not need to verify OTP for exempt views.",
        )

    def test_redirects_to_otp_verify(self):
        """The middleware should redirect to the OTP verification view if required."""
        request = self.request_factory.get("/some-path/")
        request.user = self.confirmed_staff_user

        response = self.middleware.process_view(request, basic_view, [], {})

        self.assertMiddlewareRedirects(
            response,
            "/otp/verify/?next=%2Fsome-path%2F",
            "Expected redirect to OTP verification view.",
        )

    def test_redirects_to_setup_view(self):
        """Middleware should redirect to the OTP setup view if the user has no OTP."""
        request = self.request_factory.get("/some-path/")
        request.user = self.unconfirmed_staff_user

        response = self.middleware.process_view(request, basic_view, [], {})

        self.assertMiddlewareRedirects(
            response,
            "/manage/otp/setup/?next=%2Fsome-path%2F",
            "Expected redirect to OTP setup view.",
        )



"""
Account management URLs.

Provides the URL patterns for the account management features.

Include this module in the root URL configuration:

    from hidp.accounts import account_management_urls

    urlpatterns = [
        path("manage/", include(account_management_urls)),
    ]

This module also defines the namespace `hidp_account_management` for these URLs.

Include this namespace when reversing URLs, for example:

    reverse("hidp_account_management:manage_account")
"""

from django.urls import path

from . import views

app_name = "hidp_account_management"

account_urls = [
    path("", views.ManageAccountView.as_view(), name="manage_account"),
    path("edit-account/", views.EditAccountView.as_view(), name="edit_account"),
    path(
        "edit-account/done/",
        views.EditAccountDoneView.as_view(),
        name="edit_account_done",
    ),
]

change_password_urls = [
    path(
        "change-password/",
        views.PasswordChangeView.as_view(),
        name="change_password",
    ),
    path(
        "change-password/done/",
        views.PasswordChangeDoneView.as_view(),
        name="change_password_done",
    ),
]

set_password_urls = [
    path(
        "set-password/",
        views.SetPasswordView.as_view(),
        name="set_password",
    ),
    path(
        "set-password/done/",
        views.SetPasswordDoneView.as_view(),
        name="set_password_done",
    ),
]

change_email_urls = [
    path(
        "change-email/",
        views.EmailChangeRequestView.as_view(),
        name="email_change_request",
    ),
    path(
        "change-email/sent/",
        views.EmailChangeRequestSentView.as_view(),
        name="email_change_request_sent",
    ),
    path(
        "change-email-confirm/<token>/",
        views.EmailChangeConfirmView.as_view(),
        name="email_change_confirm",
    ),
    path(
        "change-email-complete/",
        views.EmailChangeCompleteView.as_view(),
        name="email_change_complete",
    ),
    path(
        "change-email-cancel/",
        views.EmailChangeCancelView.as_view(),
        name="email_change_cancel",
    ),
    path(
        "change-email-cancel-done/",
        views.EmailChangeCancelDoneView.as_view(),
        name="email_change_cancel_done",
    ),
]

urlpatterns = (
    account_urls + change_password_urls + set_password_urls + change_email_urls
)



from datetime import timedelta
from unittest import mock

from django.contrib.auth import SESSION_KEY, get_user, get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.signals import (
    user_logged_in,
    user_logged_out,
    user_login_failed,
)
from django.http import HttpRequest
from django.test import TestCase, override_settings
from django.utils import timezone

from hidp.accounts import auth
from hidp.test.factories import user_factories


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "django.contrib.auth.backends.ModelBackend",
    ]
)
class TestAuthenticate(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()

    def setUp(self):
        self.request = HttpRequest()
        self.request.session = self.client.session

    def test_success(self):
        """
        Returns the user object if the credentials are valid.

        Does not log in the user.
        """
        user = auth.authenticate(
            request=self.request,
            email=self.user.email,
            password="P@ssw0rd!",
        )

        self.assertEqual(user, self.user)
        self.assertEqual(user.backend, "django.contrib.auth.backends.ModelBackend")
        self.assertNotIn(SESSION_KEY, self.request.session)

    @mock.patch(
        "django.contrib.auth.signals.user_login_failed.send",
        wraps=user_login_failed.send,
    )
    def test_invalid_credentials(self, mock_user_login_failed):
        """
        Returns None if the credentials are invalid.

        Then sends the `django.contrib.auth.user_login_failed` signal.
        """
        user = auth.authenticate(
            request=self.request,
            email=self.user.email,
            password="invalid",
        )

        self.assertIsNone(user)
        mock_user_login_failed.assert_called_once_with(
            sender="django.contrib.auth",
            request=self.request,
            credentials={"email": self.user.email, "password": "*" * 20},
        )

    @mock.patch(
        "django.contrib.auth.signals.user_login_failed.send",
        wraps=user_login_failed.send,
    )
    def test_permission_denied(self, mock_user_login_failed):
        """
        Returns None if the user is not allowed to log in.

        Then sends the `django.contrib.auth.user_login_failed` signal.
        """
        self.user.is_active = False
        self.user.save(update_fields=["is_active"])

        user = auth.authenticate(
            request=self.request,
            email=self.user.email,
            password="P@ssw0rd!",
        )

        self.assertIsNone(user)
        mock_user_login_failed.assert_called_once_with(
            sender="django.contrib.auth",
            request=self.request,
            credentials={"email": self.user.email, "password": "*" * 20},
        )


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "django.contrib.auth.backends.ModelBackend",
    ]
)
class TestLogin(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()

    def setUp(self):
        self.request = HttpRequest()
        self.request.session = self.client.session

    @mock.patch(
        "django.contrib.auth.signals.user_logged_in.send", wraps=user_logged_in.send
    )
    def test_success(self, mock_user_logged_in):
        """Logs in the user and sets the user in the request's session."""
        auth.login(self.request, self.user)

        self.assertEqual(self.request.session[SESSION_KEY], str(self.user.pk))
        mock_user_logged_in.assert_called_once_with(
            sender=get_user_model(),
            request=self.request,
            user=self.user,
        )

        self.user.refresh_from_db()
        self.assertAlmostEqual(
            self.user.last_login, timezone.now(), delta=timedelta(seconds=1)
        )

    @mock.patch(
        "django.contrib.auth.signals.user_logged_in.send", wraps=user_logged_in.send
    )
    def test_inactive_user(self, mock_user_logged_in):
        """Does not verify that the user is allowed to log in."""
        self.user.is_active = False
        self.user.save(update_fields=["is_active"])

        auth.login(self.request, self.user)

        self.assertEqual(self.request.session[SESSION_KEY], str(self.user.pk))
        mock_user_logged_in.assert_called_once_with(
            sender=get_user_model(),
            request=self.request,
            user=self.user,
        )

        self.user.refresh_from_db()
        self.assertAlmostEqual(
            self.user.last_login, timezone.now(), delta=timedelta(seconds=1)
        )

        # `django.contrib.auth.get_user` **does** verify that the user is
        # allowed to log in and returns an `AnonymousUser` instance.
        # This results in `request.user` being an `AnonymousUser` on the next
        # request, once `AuthenticationMiddleware` has processed the request.
        self.assertIsInstance(get_user(self.request), AnonymousUser)

    def test_none_user(self):
        """
        User may be None, and in some cases this will **not** cause an exception.

        This is unexpected behaviour, see also:
        https://code.djangoproject.com/ticket/35530#comment:1

        The wrapped `django.contrib.auth.login` function raises an exception if the user
        is not an instance of `AbstractBaseUser`.
        """
        with (
            self.subTest("request.user is absent"),
            self.assertRaisesMessage(
                TypeError, "'NoneType' does not extend AbstractBaseUser"
            ),
        ):
            auth.login(self.request, None)

        with self.subTest("Current user is None"):
            self.request.user = None
            with self.assertRaisesMessage(
                TypeError, "'NoneType' does not extend AbstractBaseUser"
            ):
                auth.login(self.request, None)

        with self.subTest("Current user is AnonymousUser"):
            self.request.user = AnonymousUser()
            with self.assertRaisesMessage(
                TypeError, "'NoneType' does not extend AbstractBaseUser"
            ):
                auth.login(self.request, None)

        # Edge case, Django uses the current user here, HIdP does not.

        with self.subTest("Current user is not None"):
            # Will log in the current user (again) for some reason.
            user = user_factories.UserFactory()
            self.request.user = user

            with self.assertRaisesMessage(
                TypeError, "'NoneType' does not extend AbstractBaseUser"
            ):
                auth.login(self.request, None)


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "django.contrib.auth.backends.ModelBackend",
    ]
)
class TestLogout(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()

    def setUp(self):
        self.request = HttpRequest()
        self.request.session = self.client.session
        self.request.user = AnonymousUser()

    @mock.patch(
        "django.contrib.auth.signals.user_logged_out.send",
        wraps=user_logged_out.send,
    )
    def test_logout_without_login(self, mock_user_logged_out):
        """Resets the session regardless of whether a user is logged in."""
        self.request.session["test"] = "test"
        session_key = self.request.session.session_key
        auth.logout(self.request)

        self.assertNotEqual(session_key, self.request.session.session_key)
        self.assertNotIn("test", self.request.session)
        mock_user_logged_out.assert_called_once_with(
            sender=type(None),  # Weird, but that's what Django does.
            request=self.request,
            user=None,
        )
        self.assertIsInstance(self.request.user, AnonymousUser)

    @mock.patch(
        "django.contrib.auth.signals.user_logged_out.send",
        wraps=user_logged_out.send,
    )
    def test_logout_after_login(self, mock_user_logged_out):
        """Logs out the user and removes the user from the request's session."""
        # Log in the user.
        auth.login(self.request, self.user)
        session_key = self.request.session.session_key
        self.assertEqual(self.request.session[SESSION_KEY], str(self.user.pk))

        # Log out the user.
        auth.logout(self.request)

        self.assertNotEqual(session_key, self.request.session.session_key)
        self.assertNotIn(SESSION_KEY, self.request.session)
        mock_user_logged_out.assert_called_once_with(
            sender=get_user_model(),
            request=self.request,
            user=self.user,
        )
        self.assertIsInstance(self.request.user, AnonymousUser)
